"""Implementation of RbaOptimization class.

Support both COBRApy and gurobipy interfaces

Note: RBA is a feasibility problem. Using a bisection algorithm as implemented in RbaPy
(Bulović et al., 2019), the highest growth rate is retrieved where the LP is still feasible.
RBA model includes variable bounds and stoichiometric coefficients that
are dependent on selected growth rate and medium.

Optimization problem related parameters need to be updated during optimization. This is performed
in according to SBML specifications using initial assignment procedures.

The class takes care of
- initial variable and constraints configuration
- extraction data configured in the SBML model, such as function definitions, initial assignments,
species references, XML annotations, Miriam annotations
- reconfiguration of optimization problem related parameters prior to each feasibility loop

Peter Schubert, HHU Duesseldorf, January 2024
"""

import re
import math
import pandas as pd
from collections import defaultdict
import tqdm


import sbmlxdf
import f2xba.prefixes as pf
from .rba_initial_assignments import InitialAssignments
from .optimize import Optimize, extract_net_fluxes
from .rba_results import RbaResults


XML_SPECIES_NS = 'http://www.hhu.de/ccb/rba/species/ns'
XML_COMPARTMENT_NS = 'http://www.hhu.de/ccb/rba/compartment/ns'


class RbaOptimization(Optimize):
    """Support optimization of SBML encoded RBA based models generated by f2xba.

    Optimization support is provided via COBRApy and guropipy interfaces for RBA models
    created by f2xba, utilizing the bisection algorithm of RbaPy (Bulović et al., 2019).

    Ref: Bulović, A., Fischer, S., Dinh, M., Golib, F., Liebermeister, W., Poirier, C., Tournier,
    L., Klipp, E., Fromion, V., & Goelzer, A. (2019). Automated generation of bacterial resource
    allocation models. Metabolic Engineering, 55, 12-22.
    https://doi.org/https://doi.org/10.1016/j.ymben.2019.06.001

    Using the COBRApy interface for optimization of RBA and TRBA models:

    .. code-block:: python

        import cobra

        cobra_model = cobra.io.read_sbml_model('iML1515_RBA.xml')
        ro = RbaOptimization('iML1515_RBA.xml', cobra_model)

        sigma = ro.avg_enz_saturation
        importer_km = ro.importer_km
        rel_mmol_per_l =  (sigma / (1.0-min(sigma, .99))) * importer_km
        ex_sidx2mmol_per_l = {re.sub('EX_', '', ex_ridx): rel_mmol_per_l for ex_ridx in lb_medium}
        ex_sidx2mmol_per_l['h_e'] = 100.0 * sigma
        ro.medium = ex_sidx2mmol_per_l

        solution = ro.solve(gr_min=0.01, gr_max=1.2, bisection_tol=1e-3)

    Using the gurobipy interface for optimization of RBA and TRBA models, providing significant speed improvements:
    Note: GUROBI optimizer with gurobipy (https://www.gurobi.com) needs to be installed on your system.

    .. code-block:: python

        ro = RbaOptimization('iML1515_RBA.xml')

        sigma = ro.avg_enz_saturation
        importer_km = ro.importer_km
        rel_mmol_per_l =  (sigma / (1.0-min(sigma, .99))) * importer_km
        ex_sidx2mmol_per_l = {re.sub('EX_', '', ex_ridx): rel_mmol_per_l for ex_ridx in lb_medium}
        ex_sidx2mmol_per_l['h_e'] = 100.0 * sigma
        ro.medium = ex_sidx2mmol_per_l

        solution = ro.solve(gr_min=0.01, gr_max=1.2, bisection_tol=1e-3)
    """

    def __init__(self, fname, cobra_model=None):
        """Instantiation of RbaOptimization instance.

        :param str fname: filename of the SBML coded extended model
        :param cobra_model: reference to cobra model (default: None)
        :type cobra_model: cobra.Model
        """
        super().__init__('RBA', fname, cobra_model)

        required = {'species', 'reactions', 'unitDefs', 'funcDefs', 'initAssign', 'parameters', 'compartments'}
        missing = required.difference(set(self.m_dict))
        if len(missing) > 0:
            print(f'Missing components {missing} in SBML document!')
            raise AttributeError

        self.df_mm_data = self._get_macromolecule_data()
        """Pandas DataFrame with macromolecule related information."""

        self.df_enz_data = self._get_enzyme_data()
        """Pandas DataFramw with enzyme molecular weights."""

        self.enz_mm_composition = self._get_enzyme_mm_composition()
        """Mapping enzyme identifiers to enzyme composition."""

        self.initial_assignments = InitialAssignments(self)
        """Initial assignments configured in the RBA model."""

        self._configure_rba_model_constraints()

        # for COBRAPy models using IBM CPLEX, switch off scaling
        if cobra_model and 'cplex' in self.model.solver.interface.__name__:
            self.model.solver.problem.parameters.read.scale.set(-1)

    @property
    def medium(self):
        """Mimic medium property of COBRApy to set and retrieve medium wrt concentrations.

        Medium concentrations returned in units of mmol/l.

        :return: metabolite concentrations in external medium in mmol/l
        :rtype: dict
        """
        ex_midx_mmol_per_l = {}
        for sid in self.ex_rid2sid.values():
            mmol_per_l = self.initial_assignments.local_env.get(sid, 0.0)
            if mmol_per_l > 0.0:
                ex_midx_mmol_per_l[re.sub(f'^{pf.M_}', '', sid)] = mmol_per_l
        return ex_midx_mmol_per_l

    @medium.setter
    def medium(self, ex_sids_mmol_per_l):
        """Mimic medium property of COBRApy for medium assignments wrt concentrations.

        Allow assignment of medium using:
          RbaOptimization.medium = ex_sids_mmol_per_l

        :param dict ex_sids_mmol_per_l: external metabolite concentrations (mmol/l)
        """
        self.set_medium_conc(ex_sids_mmol_per_l)

    def _configure_rba_model_constraints(self):
        """Configure constraints related to RBA modelling

            I.e. Enzyme forward and reverse efficiencies configured at ≤ 0.0
        """
        if self.is_gpm:
            return self._gp_configure_rba_model_constraints()
        else:
            return self._cp_configure_rba_model_constraints()

    def _gp_configure_rba_model_constraints(self):
        """Configure constraints related to RBA modelling for GuropiPy interface"""
        for constr in self.gpm.getConstrs():
            if re.match(pf.C_EF_, constr.ConstrName) or re.match(pf.C_ER_, constr.ConstrName):
                constr.sense = '<'
        print(f'RBA enzyme efficiency constraints configured (C_EF_xxx, C_ER_xxx) ≤ 0')

    def _cp_configure_rba_model_constraints(self):
        """Configure constraints related to RBA modelling for COBRApy interface"""
        modify_constr_bounds = {pf.C_EF_: [None, 0.0], pf.C_ER_: [None, 0.0]}
        for constr in self.model.constraints:
            for cprefix, bounds in modify_constr_bounds.items():
                if re.match(cprefix, constr.name):
                    constr.lb, constr.ub = bounds
        print(f'RBA enzyme efficiency constraints configured (C_EF_xxx, C_ER_xxx) ≤ 0')

    # INITIAL DATA COLLECTION PART
    def _get_macromolecule_data(self):
        mms = {}
        for sid, row in self.m_dict['species'].iterrows():
            if re.match(pf.MM_, sid):
                mm_id = re.sub(f'^{pf.MM_}', '', sid)
                refs = sbmlxdf.misc.get_miriam_refs(row.get('miriamAnnotation'), 'uniprot', 'bqbiol:is')
                uniprot = refs[0] if len(refs) > 0 else None
                xml_attrs = sbmlxdf.misc.extract_xml_attrs(row.get('xmlAnnotation'), ns=XML_SPECIES_NS)
                mw_aa = float(xml_attrs['weight_aa']) if 'weight_aa' in xml_attrs else None
                mw_kda = float(xml_attrs['weight_kDa']) if 'weight_kDa' in xml_attrs else None
                scale = float(xml_attrs['scale']) if 'scale' in xml_attrs else 1.0
                mms[mm_id] = [xml_attrs['type'], uniprot, mw_kda, mw_aa, scale]
        cols = ['type', 'uniprot', 'mw_kDa', 'mw_aa', 'scale']
        df_mms = pd.DataFrame(mms.values(), index=list(mms), columns=cols)
        df_mms.index.name = 'mm_id'
        return df_mms

    def _get_enzyme_data(self):
        """Extract molecular weights of enzymes and process machines.

        From xmlAnnotation of reactions corresponding of related concentration variables

        :return: enzyme data with name, molecular weight and scaling.
        :rtype: pandas.DataFrame
        """
        enz_data = {}
        for vid, row in self.m_dict['reactions'].iterrows():
            if re.match(pf.V_EC_, vid) or re.match(pf.V_PMC_, vid):
                xml_attrs = sbmlxdf.misc.extract_xml_attrs(row.get('xmlAnnotation'), ns=XML_SPECIES_NS)
                mw_kda = float(xml_attrs['weight_kDa']) if 'weight_kDa' in xml_attrs else None
                scale = float(xml_attrs['scale']) if 'scale' in xml_attrs else 1.0
                enz_data[vid] = [mw_kda, scale]
        cols = ['mw_kDa', 'scale']
        df_enz_data = pd.DataFrame(enz_data.values(), index=list(enz_data), columns=cols)
        df_enz_data.index.name = 'id'
        return df_enz_data

    def _get_enzyme_mm_composition(self):
        """Determine Enzyme stoichiometry wrt gene products from model stoichiometry.

        Note: stoichiometry wrt enzyme and process machine concentrations
        are configured in initial model considering a given model growth rate (e.g. 1.0 h-1)
        From stoichiometric coefficients we can derive enzyme/process machine composition
        wrt to macro molecules

        :return: enzyme/process machine composition wrt macromolecules
        :rtype: dict
        """
        model_gr = self.m_dict['parameters'].loc['growth_rate'].value
        enz_mm_composition = defaultdict(dict)
        for rid, row in self.m_dict['reactions'].iterrows():
            if re.match(pf.V_EC_, rid) or re.match(pf.V_PMC_, rid):
                for sref_str in sbmlxdf.record_generator(row['reactants']):
                    params = sbmlxdf.extract_params(sref_str)
                    if re.match(pf.MM_, params['species']):
                        mm_id = re.sub(f'^{pf.MM_}', '', params['species'])
                        enz_mm_composition[rid][mm_id] = float(params['stoic']) / model_gr
        return dict(enz_mm_composition)

    def set_medium_conc(self, ex_sids_mmol_per_l):
        """Configure external metabolite concentrations (mmol/l).

        In the RBA, the growth medium is defined by external metabolite concentrations (mmol/L).
        These concentrations are used to determine the media-dependent transporter saturation
        levels of the uptake reactions. This determination is made using a Michaelis-Menten type
        saturation function with a default Michaelis constant (Km) of 1.0 mmol/L.
        External media concentration should be scaled to the default Km value.
        This implementation of RBA still uses uptake reactions, which will be opened for
        the specified medium.

        :meta private:
        :param ex_sids_mmol_per_l: external metabolite concentrations (mmol/l)
        :type ex_sids_mmol_per_l: dict (key: metabolite id/str, val: concentration/float)
        """
        # remove any trailing 'M_'
        ex_sidx_mmol_per_l = {re.sub(f'^{pf.M_}', '', sid): mmol_per_l for sid, mmol_per_l in
                              ex_sids_mmol_per_l.items()}

        # unlimited uptake for metabolites in external medium
        self.set_medium({self.sidx2ex_ridx.get(sidx, 0.0): 1000.0 for sidx in ex_sidx_mmol_per_l})

        # configure RBA related medium concentrations (and set concentrations of not available medium to zero)
        medium_conc = {sid: ex_sidx_mmol_per_l.get(re.sub(f'^{pf.M_}', '', sid), 0.0)
                       for sid in self.ex_rid2sid.values()}
        self.initial_assignments.set_medium_conc(medium_conc)

    def _set_growth_rate(self, growth_rate):
        """Set growth rate dependent model parameters in RBA models

        :param float growth_rate: selected growth rate in h-1
        :return:
        """
        self.initial_assignments.set_growth_rate(growth_rate)

    def scale_efficiencies(self, scale_effs):
        """Scale enzyme efficiencies in RBA models to support manual parameter tuning.

        Use unscale_efficiencies() to return to original values.

        .. code-block:: python

            ro = RbaOptimization('iML1515_RBA.xml)
            scale_effs = {'PDH': {'fwd': 10.0}, 'ACt2rpp': {'fwd': 2.0, 'rev': 5.0}}
            ro.scale_efficiencies(scale_effs)
            solution = ro.solve(gr_min=0.02, gr_max=1.2, bisection_tol=1e-3)
            ro.unscale_efficiencies()

        :param dict of dict scale_effs: reaction ids (without prefix `R_`) and dict with reactions 'fwd' and/or 'rev' with float
        """
        if self.is_gpm:
            self._gp_scale_efficiencies(scale_effs)
        else:
            self._cp_scale_efficiencies(scale_effs)

    def unscale_efficiencies(self):
        """Reset previously scaled efficiencies.

        see scale_efficiencies().
        """
        if self.is_gpm:
            self._gp_unscale_efficiencies()
        else:
            self._cp_unscale_efficiencies()

    def _cp_scale_efficiencies(self, scale_effs):
        """Scale efficiencies values for COBRApy interface, by updating coupling coefficients.

        :param scale_effs: reaction identifiers (without prefix `R_`) and scaling factor
        :type scale_effs: dict (key: reaction id/str, val: dict with directions 'fwd' and/or 'rev' and factor/float)
        """
        assert self.is_gpm is False, 'applicable to COBRApy interface only'
        orig_coupling = defaultdict(dict)
        for ridx, data, in scale_effs.items():
            var_id = f'{pf.V_EC_}{ridx}'
            if var_id in self.model.reactions:
                var = self.model.reactions.get_by_id(var_id)
                for direction, factor in data.items():
                    constr_id = f'{pf.C_EF_}{ridx}' if direction == 'fwd' else f'{pf.C_ER_}{ridx}'
                    try:
                        old_val = var.get_coefficient(constr_id)
                    except KeyError:
                        print(f'Enzyme constraint not found for {ridx} in {direction} direction')
                    else:
                        orig_coupling[ridx][direction] = old_val
                        var.add_metabolites({constr_id: old_val * factor}, combine=False)
            else:
                print(f'Enzyme constraint variable not found for reaction {ridx}')
        self.orig_coupling = dict(orig_coupling)

    def _cp_unscale_efficiencies(self):
        """Unscale kcat values for COBRApy interface, by resetting original coupling coefficients
        """
        assert self.is_gpm is False, 'applicable to COBRApy interface only'
        for ridx, data, in self.orig_coupling.items():
            var = self.model.reactions.get_by_id(f'{pf.V_EC_}{ridx}')
            for direction, old_val in data.items():
                constr_id = f'{pf.C_EF_}{ridx}' if direction == 'fwd' else f'{pf.C_ER_}{ridx}'
                var.add_metabolites({constr_id: old_val}, combine=False)
        self.orig_coupling = {}

    def _gp_scale_efficiencies(self, scale_effs):
        """Scale enzyme efficiency values for gurobipy interface, by updating coupling coefficients.

        :param scale_effs: reaction identifiers (without prefix `R_`) and scaling factor
        :type scale_effs: dict (key: reaction id/str, val: dict with directions 'fwd' and/or 'rev' and factor/float)
        """
        assert self.is_gpm is True, 'applicable to gurobipy interface only'
        orig_coupling = defaultdict(dict)
        self.gpm.update()
        for ridx, data, in scale_effs.items():
            var = self.gpm.getVarByName(f'{pf.V_EC_}{ridx}')
            if var:
                for direction, factor in data.items():
                    constr_id = f'{pf.C_EF_}{ridx}' if direction == 'fwd' else f'{pf.C_ER_}{ridx}'
                    constr = self.gpm.getConstrByName(constr_id)
                    if constr:
                        old_val = self.gpm.getCoeff(constr, var)
                        orig_coupling[ridx][direction] = old_val
                        self.gpm.chgCoeff(constr, var, old_val * factor)
                    else:
                        print(f'Enzyme constraint not found for {ridx} in {direction} direction')
            else:
                print(f'Enzyme constraint variable not found for reaction {ridx}')
        self.gpm.update()
        self.orig_coupling = dict(orig_coupling)

    def _gp_unscale_efficiencies(self):
        """Reset enzyme efficiencies to original values for gurobipy models.

        Used in manually tuning model efficiencies
        - call scale_efficiencies(scale_effs) prior to optimization
        - call unscale_efficiencies() after optimization to reset old efficiency values
        """
        assert self.is_gpm is True, 'applicable to gurobipy interface only'
        for ridx, data, in self.orig_coupling.items():
            var = self.gpm.getVarByName(f'{pf.V_EC_}{ridx}')
            for direction, old_val in data.items():
                constr_id = f'{pf.C_EF_}{ridx}' if direction == 'fwd' else f'{pf.C_ER_}{ridx}'
                constr = self.gpm.getConstrByName(constr_id)
                if var and constr:
                    self.gpm.chgCoeff(constr, var, old_val)
        self.gpm.update()
        self.orig_coeffs = {}

    def get_init_assign_math(self, ia_refs):
        """Get expanded math string used in initial assignments of RBA models.

        RBA model uses initial assignments for model update during optimization.
        Model parameter values can be set depending on external media concentrations and growth rate.
        This method retrieves the expanded math for a list of initial assignment references,
        each reference contains a dictionary with the attributes:
        `var_id`: variable identifier,
        `constr_id`: constraint identifier, and
        `constr_type`: `reactants`, `products`, or `flux_bounds`.

        .. code-block:: python

            ro.get_init_assign_math([{'var_id': 'V_TSMC', 'constr_type': 'reactants', 'constr_id': 'M_hemeA_c'}])


        :param list(dict) ia_refs: reference to specific initial assignment
        :return: extended math strings
        :rtype: list(dict), like input with additional attribute 'math_str'
        """
        ia_data = []
        for data in ia_refs:
            var_id = data.get('var_id', '')
            constr_id = data.get('constr_id', '')
            constr_type = data.get('constr_type', '')

            if (var_id not in self.initial_assignments.target_rids or
                    constr_type not in ('reactants', 'products', 'flux_bounds') or
                    constr_id not in getattr(self.initial_assignments.target_rids[var_id], constr_type)):
                print(f'Invalid reference. var_id: {var_id}, constr_type: {constr_type}, constr_id: {constr_id}')
            else:
                trid = self.initial_assignments.target_rids[var_id]
                symbol_id = getattr(trid, constr_type)[constr_id]
                ia_func = self.initial_assignments.ia_functions[symbol_id]
                ia_data.append({'var_id': var_id, 'constr_type': constr_type, 'constr_id': constr_id,
                                 'math_str': ia_func.expanded_math})
        return ia_data

    def set_init_assign_math(self, ia_data):
        """Set expanded math string used in initial assignments of RBA models.

         RBA model uses initial assignments for model update during optimization.
         Model parameter values can be set depending on external media concentrations and growth rate.
         This method sets the expanded math for a list of initial assignment references,
         each reference contains a dictionary with the attributes:
         `var_id`: variable identifier,
         `constr_id`: constraint identifier,
         `constr_type`: `reactants`, `products`, or `flux_bounds`, and
         `math`: math string, e.g. `growth_rate * 0.002` or a fixed value as 0.0.

        .. code-block:: python

            ia_data_anaerobe = [{'var_id': 'V_TSMC', 'constr_type': 'reactants', 'constr_id': 'M_hemeA_c', 'math_str': '0.0'}]
            if 'EX_o2_e' not in medium:
                orig_ia_data = ro.set_init_assign_math(ia_data_anaerobe)
            solution = ro.solve(gr_min=0.01, gr_max=0.7, bisection_tol=1e-3)
            if 'EX_o2_e' not in medium:
                ro.set_init_assign_math(orig_ia_data)

         :param list(dict) ia_data: reference to specific initial assignment
         :return: the original math string, prior to modification
         :rtype: list(dict), like input, but with original math string
         """
        orig_ia_data = []
        for data in ia_data:
            var_id = data.get('var_id', '')
            constr_id = data.get('constr_id', '')
            constr_type = data.get('constr_type', '')
            new_math = str(data.get('math_str', ''))

            if (var_id not in self.initial_assignments.target_rids or
                    constr_type not in ('reactants', 'products', 'flux_bounds') or
                    constr_id not in getattr(self.initial_assignments.target_rids[var_id], constr_type) or
                    len(new_math) == 0):
                print(
                    f'Invalid record. var_id: {var_id}, constr_type: {constr_type}, constr_id: {constr_id}, math_str: {new_math}')
            else:
                trid = self.initial_assignments.target_rids[var_id]
                symbol_id = getattr(trid, constr_type)[constr_id]
                ia_func = self.initial_assignments.ia_functions[symbol_id]
                org_math = ia_func.expanded_math
                ia_func.expanded_math = new_math
                orig_ia_data.append(
                    {'var_id': var_id, 'constr_type': constr_type, 'constr_id': constr_id, 'math_str': org_math})
        return orig_ia_data

    def solve(self, gr_min=0.0, gr_max=1.5, bisection_tol=1e-5, max_iter=40):
        """Solve RBA feasibility problem using bisection algorithm.

        Support both COBRApy and gurobipy interfaces.

        :param float gr_min: minimum growth rate in h-1
        :param float gr_max: maximum growth rate in h-1
        :param float bisection_tol: stop criteria based on growth rate tolerances (default: 1e-3)
        :param int max_iter:  stop criteria based on iteration number (default: 40)
        :return: optimization solution
        :rtype: class:`Solution`
        """
        if self.is_gpm:
            return self._gp_solve(gr_min, gr_max, bisection_tol, max_iter)
        else:
            return self._cp_solve(gr_min, gr_max, bisection_tol, max_iter)

    def _gp_solve(self, gr_min, gr_max, bisection_tol, max_iter):
        """Solve RBA feasibility problem using bisection algorithm of RbaPy 1.0.

        Optionally we can modify the model to non-negative variables only.

        :param float gr_min: minimum growth rate to test in h-1, might have to be > 0.0 depending on targets
        :param float gr_max: maximum growth rate to test in h-1
        :param float bisection_tol: (optional, default 1e-3) stop criteria based on growth rate tolerances
        :param int max_iter: (optional, default 40) stop criteria based on number of iterations
        :return: optimization solution
        :rtype: class:`Solution`
        """
        # bisection algorithm to narrow in on maximum growth rate
        solution = None
        self._set_growth_rate(gr_min)
        if self.optimize().status == 'optimal':
            n_iter = 1
            while ((gr_max - gr_min) > bisection_tol) and (n_iter < max_iter):
                gr_test = gr_min + 0.5 * (gr_max - gr_min)
                self._set_growth_rate(gr_test)
                # self.gpm.reset()
                if self.optimize().status == 'optimal':
                    gr_min = gr_test
                else:
                    gr_max = gr_test
                n_iter += 1

            # collect optimization solution at optimum growth rate
            if (gr_max - gr_min) < bisection_tol:
                gr_opt = gr_min
                self._set_growth_rate(gr_opt)
                solution = self.optimize()

                # if solution with gr_min is no longer optimal, lower gr_opt slightly and try again
                if solution.status != 'optimal':
                    gr_opt = max(0.0, gr_opt - 0.5 * bisection_tol)
                    self._set_growth_rate(gr_opt)
                    solution = self.optimize()

                solution.fluxes = extract_net_fluxes(solution.fluxes)
                solution.objective_value = gr_opt
                solution.n_iter = n_iter
                # determine final density constraint
                col = self.gpm.getCol(self.gpm.getVarByName('V_TCD'))
                solution.density_constraints = {col.getConstr(idx).constrname: -col.getCoeff(idx)
                                                for idx in range(col.size())}
            else:
                print('no optimal growth rate')

        return solution

    def _cp_solve(self, gr_min, gr_max, bisection_tol, max_iter):
        """Solve RBA feasibility problem using bisection algorithm of RbaPy 1.0.

        Use COBRApy model contexts, so we can support outer model contexts

        :param float gr_min: minimum growth rate to test in h-1, might have to be > 0.0 depending on targets
        :param float gr_max: maximum growth rate to test in h-1
        :param float bisection_tol: (optional, default 1e-3) stop criteria based on growth rate tolerances
        :param int max_iter: (optional, default 40) stop criteria based on number of iterations
        :return: optimization solution
        :rtype: class:`Solution`
        """
        # first, check feasibility a minimal (or zero) growth
        # with self.model:
        self._set_growth_rate(gr_min)
        opt_value = self.model.slim_optimize()

        # bisection algorithm to narrow in on maximum growth rate
        if math.isfinite(opt_value):
            n_iter = 1
            while ((gr_max - gr_min) > bisection_tol) and (n_iter < max_iter):
                gr_test = gr_min + 0.5 * (gr_max - gr_min)
                # with self.model:
                self._set_growth_rate(gr_test)
                opt_value = self.model.slim_optimize()
                if math.isfinite(opt_value):
                    gr_min = gr_test
                else:
                    gr_max = gr_test
                n_iter += 1

            # collect optimization solution at optimum growth rate
            if (gr_max - gr_min) < bisection_tol:
                gr_opt = gr_min
                self._set_growth_rate(gr_opt)
                solution = self.model.optimize()

                # if solution with gr_min is no longer optimal, lower gr_opt slightly and try again
                if solution.status != 'optimal':
                    gr_opt = max(0.0, gr_opt - 0.5 * bisection_tol)
                    solution = self.model.optimize()

                solution.objective_value = gr_opt
                solution.n_iter = n_iter
                solution.density_constraints = {met.id: -val for met, val
                                                in self.model.reactions.get_by_id(pf.V_TCD).metabolites.items()}
                return solution
            else:
                print('no optimal growth rate')
                return None
        else:
            print(f'Problem infeasible at minimum growth of {gr_min}')
            return None

    def single_gene_deletion(self, genes=None, solution=None, **kwargs):
        """Perform a single gene deletion analysis for RBA based models using gurobipy interface.

        Interface aligned to COBRApy single_gene_deletion() method.
        Perform single gene deletion simulations for provided list of gene in `genes`.
        If `genes` is not provided, perform gene deletion for all genes that may be
        active in the wild type solution. In case a gene is not required in the wild type solution, a knockout
        simulation is not performed for this gene. Its growth rate value is set to the wild type value and
        its optimization status is set to `wt_solution`.
        A wild type solution can be provided with parameter `solution`, alternatively a wild type solution
        is determined automatically.

        Following keyword arguments can be added to the list of parameters:
        `gr_min`: (default 0.0) minimum growth rate to test in h-1,
        `gr_min_alt`: (optional) alternative minimum growth rate in h-1 (used, if infeasible solution with gr_min),
        `gr_max`: (default 1.5) maximum growth rate to test in h-1,
        `bisection_tol`: (default 1e-5) stop criteria based on growth rate tolerance,
        `max_iter`: (default 40) stop criteria based on number of iterations.

        :param list or set genes: (optional) gene ids,
        :param solution: (optional) wild type RBA solution
        :type solution: :class:`Solution`
        :param kwargs: keyword arguments used for RBA bisectional optimization
        :return: single gene deletion results per gene with growth rate in h-1, optimization status and fitness value
        :rtype: pandas.DataFrame
        """
        if self.is_gpm is None:
            print('Method implemented for gurobipy interface only.')
            return pd.DataFrame()

        alt_kwargs = kwargs.copy()
        if 'gr_min_alt' in kwargs:
            alt_kwargs['gr_min'] = kwargs.pop('gr_min_alt')
            del alt_kwargs['gr_min_alt']

        # determine wild type growth rate and fluxes, if solution not provided
        if solution is None:
            solution = self.solve(**kwargs)
        wt_gr = solution.objective_value
        wt_fluxes = dict(RbaResults(self, {'wt': solution}).collect_fluxes()['wt'])

        all_genes = set(self.m_dict['fbcGeneProducts'].label.values)
        tx_genes, metab_genes = self.get_tx_metab_genes()
        pm_genes = all_genes.difference(tx_genes.union(metab_genes))

        # determine gene list, if not provided
        if genes is None:
            genes = self.get_active_genes(wt_fluxes)
        else:
            all_genes = genes

        # optimization loop for single gene deletions
        cols = ['growth_rate', 'status', 'fitness']
        sgko_results = {'wt': [wt_gr, 'wt_solution', 1.0]}
        for gene in tqdm.tqdm(sorted(all_genes)):
            if gene in genes:
                # simulate a single gene deletion
                orig_rid_bounds = self.gene_knock_outs(gene)
                solution = self.solve(**kwargs)
                if (solution is not None) and solution.status == 'infeasible':
                    print(f'first solution infeasible - reset model {gene}')
                    self.gpm.reset()
                    solution = self.solve(**alt_kwargs)
                    print(f'solution status second attempt for {gene}: {solution.status}')
                self.set_variable_bounds(orig_rid_bounds)

                # process simulation result
                if solution is None:
                    sgko_results[gene] = [0.0, 'infeasible', 0.0]
                elif solution.status in {'optimal'}:
                    mutant_gr = solution.objective_value
                    sgko_results[gene] = [mutant_gr, solution.status, mutant_gr / wt_gr]
                else:
                    sgko_results[gene] = [0.0, solution.status, 0.0]
            elif gene in pm_genes:
                sgko_results[gene] = [0.0, 'process_machine']
            else:
                # genes not in gene_list are assumed to not impact the wild type solution
                sgko_results[gene] = [wt_gr, 'wt_solution', 1.0]

        df_sgko = pd.DataFrame(sgko_results.values(), index=list(sgko_results), columns=cols)
        df_sgko.index.name = 'gene'

        return df_sgko



"""Implementation of FbaOptimization class.

Support functions for COBRApy and gurobipy optimization of FBA models.

Peter Schubert, HHU Duesseldorf, CCB, November 2024
"""

import re
import pandas as pd
import tqdm

# gurobipy should not be a hard requirement, unless used in this context
try:
    import gurobipy as gp
except ImportError:
    gp = None
    pass

from .optimize import Optimize
from .fba_results import FbaResults


class FbaOptimization(Optimize):
    """Support optimization of SBML encoded FBA based models generated by f2xba.

    Using the COBRApy interface for optimization of FBA/TFA models:
    Use of FbaOptimization is optional, but it can provide access to additional features, e.g. results analysis,
    and ensures correct configuration of variables and constraints for thermodynamics constraint variants.

    .. code-block:: python

        import cobrapy

        cobra_model = cobra.io.read_sbml_model('iML1515.xml')
        fo = FbaOptimization('iML1515.xml', cobra_model)
        cobra_model.medium = lb_medium
        solution = cobra_model.optimize()

    Using the gurobipy interface for optimization of FBA/TFA models:
    Note: GUROBI optimizer with gurobipy (https://www.gurobi.com) needs to be installed on your system.

    .. code-block:: python

        fo = FbaOptimization('iML1515.xml')
        fo.medium = lb_medium
        solution = fo.optimize()
    """

    def __init__(self, fname, cobra_model=None):
        """Instantiate the FbaOptimization instance.

        :param str fname: filename of the SBML coded FBA/TFA model
        :param cobra_model: reference to cobra model (default: None)
        :type cobra_model: :class:`cobra.Model`
        """
        super().__init__('FBA', fname, cobra_model)

    @property
    def medium(self):
        """Mimic medium property of COBRApy to set and retrieve medium.

        :return: exchange reaction ids with (positive valued) uptake rates
        :rtype: dict
        """
        ex_medium = {}
        for ex_rid, (lb, ub) in self.get_variable_bounds(self.uptake_rids).items():
            if lb < 0.0:
                ex_medium[re.sub('^R_', '', ex_rid)] = -lb
        return ex_medium

    @medium.setter
    def medium(self, ex_medium):
        """Mimic medium property of COBRApy for medium assignments.

        Allow assignment of medium using:
          FbaOptimization.medium = ex_medium

        :param dict ex_medium: exchange reaction ids (with/without `R_`) and (positive valued) uptake rates
        """
        self.set_medium(ex_medium)

    def single_gene_deletion(self, genes=None, method='fba', solution=None, **kwargs):
        """Perform a single gene deletion analysis for FBA based models using gurobipy interface.

        Interface aligned to COBRApy single_gene_deletion() method.
        Perform single gene deletion simulations for provided list of gene in `genes`.
        If `genes` is not provided, perform gene deletion for all genes that may be
        active in the wild type solution. In case a gene is not required in the wild type solution, a knockout
        simulation is not performed for this gene. Its growth rate value is set to the wild type value and
        its optimization status is set to `wt_solution`.
        With `method` set to `moma`, `linear moma`, `room` or `linear room`, respective method is used for
        the optimization.
        A wild type solution can be provided with parameter `solution`, alternatively a wild type solution
        is determined automatically.

        When `method` is set to `room` or `linear room`, following keyword arguments can be added:
        `delta`: relative tolerance range (default: 0.03),
        `epsilon`: absolute tolerance range (default: 1e-3),
        `time_limit`: in seconds for single gene deletion simulation, used for 'room' (default: 30.0).

        .. code-block:: python

            fo = FbaOptimization('iML1515.xml')
            fo.medium = lb_medium
            df_sgko = fo.single_gene_deletion()

        Example for MOMA based SGKO analysis for selected genes with wild type solution provided:

        .. code-block:: python

            pfba_solution = fo.pfba()
            fo.single_gene_deletion(genes=['b0002', 'b0003', 'b0007', 'b0025'], method='moma', solution=pfba_solution)

        :param list or set genes: (optional) gene ids,
        :param str method: (optional) alternative methods 'moma', 'linear moma', 'room' or 'linear room'
        :param solution: (optional) wild type FBA solution
        :type solution: :class:`Solution`
        :param kwargs: keyword arguments passed on to 'room' and 'linear room' methods
        :return: Table with SGKO results, containing growth rate in h-1, optimization status and fitness
        :rtype: pandas.DataFrame
        """
        if self.is_gpm is None:
            print('Method implemented for gurobipy interface only.')
            return pd.DataFrame()

        linear = True if 'linear' in method else False
        biomass_rid = self._get_biomass_rid()

        # determine wild type growth rate and fluxes, if solution not provided
        # extract wt_gr from fluxes of biomass_rid (assume growth maximization of Biomass objective function)
        if solution is None:
            solution = self.optimize()
        wt_gr = solution.fluxes[biomass_rid]
        wt_fluxes = dict(FbaResults(self, {'wt': solution}).collect_fluxes()['wt'])

        # determine gene list, if not provided
        if genes is None:
            genes = self.get_active_genes(wt_fluxes)
            all_genes = self.m_dict['fbcGeneProducts'].label.values
        else:
            all_genes = genes

        # optimization loop for single gene deletions
        cols = ['growth_rate', 'status', 'fitness']
        sgko_results = {'wt': [wt_gr, 'wt_solution', 1.0]}
        for gene in tqdm.tqdm(sorted(all_genes)):
            if gene in genes:

                # simulate a single gene deletion
                orig_rid_bounds = self.gene_knock_outs(gene)
                if 'moma' in method:
                    solution = self.moma(wt_fluxes, linear=linear)
                elif 'room' in method:
                    solution = self.room(wt_fluxes, linear=linear, **kwargs)
                else:
                    solution = self.optimize()
                self.set_variable_bounds(orig_rid_bounds)

                # process simulation result
                if solution is None:
                    sgko_results[gene] = [0.0, 'infeasible', 0.0]
                elif solution.status in {'optimal', 'time_limit', 'suboptimal'}:
                    if 'moma' in method or 'room' in method:
                        mutant_gr = solution.fluxes[biomass_rid]
                    else:
                        mutant_gr = solution.objective_value
                    sgko_results[gene] = [mutant_gr, solution.status, mutant_gr / wt_gr]
                else:
                    sgko_results[gene] = [0.0, solution.status, 0.0]
            else:
                # genes not in gene_list are assumed to not impact the wild type solution
                sgko_results[gene] = [wt_gr, 'wt_solution', 1.0]

        df_sgko = pd.DataFrame(sgko_results.values(), index=list(sgko_results), columns=cols)
        df_sgko.index.name = 'gene'

        return df_sgko



"""Implementation of RbaModel class used in f2xba.

Extend the XbaModel to become a resource balance constraint model (RBA).
RBA implementation is based on RbaPy (Bulović et al., 2019)

Peter Schubert, CCB, HHU Duesseldorf, December 2022
"""

import os
import re
import pandas as pd
from collections import defaultdict

from .rba_macromolecules import RbaMacromolecules
from .rba_metabolism import RbaMetabolism
from .rba_parameters import RbaParameters, RbaFunction
from .rba_processes import RbaProcesses
from .rba_enzymes import RbaEnzymes
from .rba_densities import RbaDensities
from .rba_targets import RbaTargets
from .rba_medium import RbaMedium
from ..utils.calc_mw import protein_mw_from_aa_comp, rna_mw_from_nt_comp, ssdna_mw_from_dnt_comp, get_seq_composition
from ..utils.mapping_utils import valid_sbml_sid, load_parameter_file
from .initital_assignments import InitialAssignments
import f2xba.prefixes as pf

DEFAULT_GROWTH_RATE = 1.0
"""Growth rate used to configure growth rate dependent parameters in the SBML model."""

DEFAULT_ENZ_SATURATION = 0.5
"""Default enzyme saturation level, unless adjusted by `avg_enz_sat` in table `general`."""

DEFAULT_MICHAELIS_CONSTANT = 1.0
"""Default Michaelis constant KM for medium uptake reactions in mmol/l"""

MAX_ENZ_CONC = 10
"""Upper bound enzyme concentration in mmol/gDW"""

MAX_MM_PROCESSING_FLUX = 10
"""Upper bound for macromolecular synthesis and degradation fluxes in µmol/gDWh"""

MAX_DENSITY_SLACK = 10
"""Upper bound on compartmental density in mmolAA/gDW"""

XML_SPECIES_NS = 'http://www.hhu.de/ccb/rba/species/ns'
XML_REACTION_NS = 'http://www.hhu.de/ccb/rba/reaction/ns'
XML_COMPARTMENT_NS = 'http://www.hhu.de/ccb/rba/compartment/ns'

components = {'parameters', 'dna', 'rnas', 'proteins', 'metabolism',
              'processes', 'enzymes', 'densities', 'targets', 'medium'}


class RbaModel:
    """Create a resource balance constraint model (RBA or TRBA) by extending a XbaModel instance.

    RBA mplementation is based on the Python package RbaPy (Bulović et al., 2019).

    Ref: Bulović, A., Fischer, S., Dinh, M., Golib, F., Liebermeister, W., Poirier, C., Tournier,
    L., Klipp, E., Fromion, V., & Goelzer, A. (2019). Automated generation of bacterial resource
    allocation models. Metabolic Engineering, 55, 12-22.
    https://doi.org/https://doi.org/10.1016/j.ymben.2019.06.001

    Example: Create a RBA model by extending an existing genome scale metabolic model.
    The spreadsheet files `xba_parameters.xlsx` and `rba_parameters.xlsx`
    contain configuration data. See tutorials.

    .. code-block:: python

        xba_model = XbaModel('iML1515.xml')
        xba_model.configure('xba_parameters.xlsx')

        rba_model = RbaModel(xba_model)
        rba_model.configure('rba_parameters.xlsx')
        if rba_model.validate():
            rba_model.export('iML1515_RBA.xml')
    """

    def __init__(self, xba_model):
        """Instantiate the RbaModel instance.

        :param xba_model: a reference to a XbaModel instance
        :type xba_model: :class:`f2xba.XbaModel`
        """
        self.model = xba_model
        """Reference to the XbaModel instance."""

        self.dna = RbaMacromolecules('dna')
        """Reference to DNA macromolecule configuration."""

        self.rnas = RbaMacromolecules('rna')
        """Reference to RNA macromolecule configurations."""

        self.proteins = RbaMacromolecules('protein')
        """Reference to protein macromolecule configurations."""

        self.metabolism = RbaMetabolism()
        """Reference to RBA compartment, species and reaction configurations."""

        self.parameters = RbaParameters()
        """Reference to RBA function configurations."""

        self.processes = RbaProcesses()
        """Reference to RBA processes configurations."""

        self.densities = RbaDensities()
        """Reference to RBA compartment density configurations."""

        self.targets = RbaTargets()
        """Reference to RBA targets configurations."""

        self.enzymes = RbaEnzymes()
        """Reference to RBA enzyme configurations."""

        self.medium = RbaMedium()
        """Reference to RBA medium configuration."""

        self.cid_mappings = {}
        """Map compartment identifiers to compartment specific roles."""

        self.mmid2mm = {}
        """Map macromolecules to configuration data."""

        self.initial_assignments = InitialAssignments(self.parameters)
        """Reference to initial assignments used in the SBML model."""

        self.parameter_values = {}
        """Parameter values configured in the model."""

        self.avg_enz_sat = None

    def _get_dummy_translation_targets(self, rba_params):
        """Extract translation targets for dummy proteins.

        Extract translation targets from rba_params['compartments'] and
        return target data for dummy protein translation targets

        :param dict(str, pandas.DataFrame) rba_params: rba configuration data loaded from file
        :return: target configuration data related to dummy proteins
        :rtype: pandas.DataFrame
        """
        df_c_data = rba_params['compartments']
        dummy_proteins = self.cid_mappings['dummy_proteins']

        df = df_c_data.loc[dummy_proteins.values()]
        df.rename(columns=lambda x: re.sub(r'^translation_', '', x), inplace=True)
        df['target_group'] = 'translation_targets'
        df['target_type'] = 'concentrations'
        df['target'] = [f'dummy_protein_{c_name}' for c_name in df.index]
        if 'target_value_type' not in df.columns:
            df['target_value_type'] = 'value'

        cols = [col for col in df.columns if re.match('target', col)]
        df = df[cols]
        df.set_index('target_group', inplace=True)
        return df

    def _add_dummy_proteins(self):
        """Add dummy proteins to different compartments based on configuration.

        Use average protein composition in terms of amino acids.
        """
        dummy_proteins = self.cid_mappings['dummy_proteins']

        # determine average protein composition
        aa_count = defaultdict(int)
        for p in self.model.proteins.values():
            for cmp_id, count in get_seq_composition(p.aa_sequence).items():
                aa_count[cmp_id] += count
        total_count = sum(aa_count.values())
        avg_aa_len = total_count / len(self.model.proteins)
        composition = {aa: avg_aa_len * count / total_count for aa, count in aa_count.items()}

        # Amino acids with low frequency (here, selenocysteine), set stoic coefficient to zero
        for aa in composition:
            if composition[aa] < 0.01:
                composition[aa] = 0.0

        for p_name, cid in dummy_proteins.items():
            self.proteins.add_macromolecule(p_name, cid, composition)

        # update average protein length parameter (assuming this parameter is used for dummy targets)
        f_name = 'inverse_average_protein_length'
        self.parameters.functions[f_name] = RbaFunction(f_name, f_type='constant',
                                                        f_params={'CONSTANT': 1.0 / avg_aa_len})

        print(f'{len(dummy_proteins):4d} dummy proteins added')

    @staticmethod
    def _get_cid_mappings(rba_params):
        """Extract specific compartment functions.

        Depending on 'type' column in sheet 'compartments'
        - 'cytoplasm': default compartment for unassigned gene products, e.g. rRNA proteins
        - 'uptake': compartment with transporters for medium uptake, e.g. 'outer_membrane'
        - 'medium': compartment in which species for uptake are located, e.g. 'external'

        Mapping of XbaModel compartment ids to RBA compartment names
        Each reaction is connected to a compartment (based on its substrates)
        Transport reactions are connected to a compartment composed of a sorted concatenation
        of substrate cids, e.g. 'c-p'.

        returns model specific compartment id mappings required for RBA model. A dict with keywords
        - rcid2cid: mapping of reaction CIDs to CIDs, dict
        - uptake_rcids: reaction CIDs for transport reactions involved in medium uptake, set of str
        - medium_cid: CID where medium is located, str
        - dummy_proteins: name of dummy proteins and their CID location, dict

        e.g.  {'rcid2cid': {'c': 'c', 'e': 'e', 'p': 'p', 'c-p': 'im', 'c-e-p': 'im', 'c-e': 'im', 'e-p': 'om'},
               'uptake_rcids': {'e-p'},
               'cytoplasm_cid': 'c',
               'medium_cid': 'e',
               'dummy_proteins': {'dummy_protein_c': 'c', 'dummy_protein_e': 'e', 'dummy_protein_p': 'p',
               'dummy_protein_im': 'im', 'dummy_protein_om': 'om'}}

        :param dict(str, pandas.DataFrame) rba_params: RBA specific parameters loaded from parameter Excel spreadsheet
        :return: model specific compartment id mappings
        :rtype: dict
        """
        df_c_data = rba_params['compartments']

        # map compartment ids (cids) to reaction cids - mainly required for membrane compartments
        cid_mappings = {}
        cid2rcids = {}
        for cid, row in df_c_data.iterrows():
            if type(row['reaction_cids']) is str:
                cid2rcids[cid] = {rcid.strip() for rcid in row['reaction_cids'].split(',')}

        cid_mappings['rcid2cid'] = {}
        for cid, rcids in cid2rcids.items():
            for rcid in rcids:
                cid_mappings['rcid2cid'][rcid] = cid

        cid_mappings['uptake_rcids'] = set()
        for cid, row in df_c_data.iterrows():
            if row['keyword'] == 'uptake':
                cid_mappings['uptake_rcids'] |= cid2rcids[cid]
            elif row['keyword'] == 'medium':
                cid_mappings['medium_cid'] = cid
            elif row['keyword'] == 'cytoplasm':
                cid_mappings['cytoplasm_cid'] = cid
            elif row['keyword'] == 'total_density':
                cid_mappings['total_density_cid'] = cid

        # identify compartments with translation targets for dummy proteins
        cid_mappings['dummy_proteins'] = {}
        for col in df_c_data.columns:
            if re.match('translation_target', col):
                cids = df_c_data[df_c_data[col].notna()].index
                for cid in cids:
                    p_name = f'dummy_protein_{cid}'
                    if p_name not in cid_mappings['dummy_proteins']:
                        cid_mappings['dummy_proteins'][p_name] = cid
        return cid_mappings

    def _prepare_xba_model(self, rba_params):
        """Prepare XbaModel instance for conversion to RBA.

        extract data from:
            rba_params['machineries']
            rba_params['compartments']

        Add membrane compartment to the XbaModel based on 'compartments' sheet.
        Add gene products required for RBA machineries based on 'machineries' sheet.
        Create relevant proteins in the XbaModel instance and map cofactors.
        Split reactions into (reversible) iso-reactions, each catalyzed by a single enzyme.

        :param dict(str, pandas.DataFrame) rba_params: RBA specific parameters loaded from parameter spreadsheet
        """
        # add (membrane) compartments to the XbaModel
        cs_config = {}
        for cid, row in rba_params['compartments'].iterrows():
            if cid not in self.model.compartments:
                cs_config[cid] = {'name': row['name']}
        self.model.add_compartments(cs_config)
        if len(cs_config) > 0:
            print(f'{len(cs_config):4d} compartment(s) added')

        # add gene products for coenzymes
        if 'coenzymes' in rba_params:
            count = self.model.add_gps(rba_params['coenzymes'].set_index('gpid', drop=False))
            print(f'{count:4d} gene product(s) added for coenzymes')
            self.model.update_gp_mappings()

        # add gene products for process machineries
        df_mach_data = rba_params['machineries']
        count = self.model.add_gps(df_mach_data[df_mach_data['set'] == 'protein'].set_index('gpid'))
        if count > 0:
            print(f'{count:4d} gene product(s) added for process machineries')
            self.model.update_gp_mappings()

        # create model proteins for newly added gene products
        count = self.model.create_proteins()
        if count > 0:
            print(f'{count:4d} protein(s) created with UniProt information')

        if self.model.cofactor_flag is True:
            count = self.model.map_protein_cofactors()
            if count > 0:
                print(f'{count:4d} cofactor(s) mapped to species ids for added protein')

        # split reactions into (reversible) iso-reactions
        n_r = len(self.model.reactions)
        self.model.add_isoenzyme_reactions()
        n_isor = len(self.model.reactions)
        print(f'{n_r} reactions -> {n_isor} iso-reactions, including pseudo reactions')

    def configure(self, fname):
        """Configure the RbaModel instance with information provided in the RBA configuration file.

        The RBA configuration spreadsheet must contain following sheets:
        'general', 'trna2locus', 'compartments', 'targets', 'functions', 'processing_maps',
        'processes' and 'machineries'. The sheet 'coenzymes' is optional.

        Example: Create a RBA model by extending an existing genome scale metabolic model.
        The spreadsheet files `xba_parameters.xlsx` and `rba_parameters.xlsx`
        contain configuration data. See tutorials.

        .. code-block:: python

            xba_model = XbaModel('iML1515.xml')
            xba_model.configure('xba_parameters.xlsx')

            rba_model = RbaModel(xba_model)
            rba_model.configure('rba_parameters.xlsx')

        :param str fname: filename of RBA configuration file (.xlsx)
        :return: success status
        :rtype: bool
        """
        sheet_names = ['general', 'trna2locus', 'coenzymes', 'compartments', 'targets',
                       'functions', 'processing_maps', 'processes', 'machineries']
        required_sheets = {'general', 'trna2locus', 'compartments', 'targets',
                           'functions', 'processing_maps', 'processes', 'machineries'}
        rba_params = load_parameter_file(fname, sheet_names)
        missing_sheets = required_sheets.difference(set(rba_params.keys()))
        if len(missing_sheets) > 0:
            print(f'missing required tables {missing_sheets}')
            raise ValueError

        if not self._check_rba_params_functions(rba_params):
            print('ERRORs in RBA Parameter file')
            return False
        if not self._check_rba_params_labels(rba_params):
            print('ERRORs in RBA Parameter file')
            return False

        general_params = rba_params['general']['value'].to_dict()
        self.avg_enz_sat = general_params.get('avg_enz_sat', DEFAULT_ENZ_SATURATION)

        self._prepare_xba_model(rba_params)

        self.cid_mappings = self._get_cid_mappings(rba_params)
        # check compartment mapping
        used_rcids = set()
        for p in self.model.proteins.values():
            used_rcids.add(p.compartment)
        ok_flag = True
        for rcid in used_rcids:
            if rcid not in self.cid_mappings['rcid2cid']:
                print(f'no mapping of "{rcid}" to RBA compartment. Assign mapping in RBA compartments table!')
                ok_flag = False
        if not ok_flag:
            print('ERRORs in RBA Parameter file')
            return False

        self.parameters.from_xba(rba_params)
        self.densities.from_xba(rba_params, self.parameters)
        self.targets.from_xba(rba_params, self.model, self.parameters)
        df_dummy_targets = self._get_dummy_translation_targets(rba_params)
        self.targets.from_xba({'targets': df_dummy_targets}, self.model, self.parameters)
        self.metabolism.from_xba(rba_params, self.model)

        self.medium.from_xba(general_params, self.model)

        macromolecules = set(rba_params['processes']['set'].values)
        if 'dna' in macromolecules:
            self.dna.from_xba(rba_params, self.model, self.cid_mappings)
        if 'rna' in macromolecules:
            self.rnas.from_xba(rba_params, self.model, self.cid_mappings)
        self.proteins.from_xba(rba_params, self.model, self.cid_mappings)
        self._add_dummy_proteins()
        self.enzymes.from_xba(self.avg_enz_sat, self.model, self.parameters, self.cid_mappings, self.medium,
                              DEFAULT_MICHAELIS_CONSTANT)
        self.processes.from_xba(rba_params, self)

        print(f'{len(self.parameters.functions):4d} functions, {len(self.parameters.aggregates):4d} aggregates')
        print(f'>>> RBA model created')

        self._update_xba_model()
        return True

    def _update_xba_model(self):
        """Based on configured RBA model update the corresponding XbaModel.

        RbaModel can be exported in RBA proprietary format.
        XbaModel can be exported in standardized SBML format.

        values of growth_rate and configured medium are used to calculate
        function values, like enzyme efficiencies, enzyme/process concentration
        requirements, targets

        Note: some XbaModel updates already performed in _prepare_xba_model()
        """
        print('update XbaModel with RBA parameters')

        growth_rate = DEFAULT_GROWTH_RATE
        protect_ids = set()

        # create units required for RBA
        rba_units = {'hour': {'name': 'hour', 'units': 'kind=second, exp=1.0, scale=0, mult=3600.0'},
                     'per_h': {'name': 'per hour', 'units': 'kind=second, exp=-1.0, scale=0, mult=3600.0'},
                     'mmol_per_gDW': {'name': 'Millimoles', 'units': 'kind=mole, exp=1.0, scale=-3, mult=1.0; '
                                                                     'kind=gram, exp=-1.0, scale=0, mult=1.0'}
                     }
        for unit_id, data in rba_units.items():
            if unit_id not in self.model.unit_defs:
                data['id'] = unit_id
                self.model.add_unit_def(data)
            protect_ids.add(unit_id)

        pid = 'growth_rate'
        p_data = {'value': growth_rate, 'units': 'per_h', 'name': 'growth rate for initial parametrization'}
        self.model.add_parameter(pid, p_data)
        protect_ids.add(pid)

        # identify compartment where RBA medium is located
        xml_annot = f'ns_uri={XML_COMPARTMENT_NS}, prefix=rba, token=compartment, medium=True'
        add_annot = [['compartment', 'xml_annotation', xml_annot]]
        cols = ['component', 'attribute', 'value']
        df_modify_attrs = pd.DataFrame(add_annot, columns=cols, index=[self.cid_mappings['medium_cid']])
        self.model.modify_attributes(df_modify_attrs, 'compartment')

        # mapping of macromolecule ids to RbaMacromolecule instances
        # TODO: do we actually need self.mmid2mm ?
        mm_types = {'protein': self.proteins, 'dna': self.dna, 'rna': self.rnas}
        for mm_type, mm_data in mm_types.items():
            for mm_id, mm in mm_data.macromolecules.items():
                self.mmid2mm[mm_id] = mm

        # remove biomass reactions for the model
        rids = [rid for rid, r in self.model.reactions.items() if r.kind == 'biomass']
        self.model.del_components('reactions', rids)

        self._xba_add_macromolecule_species()
        self._xba_add_rba_constraints()
        self._xba_couple_reactions()
        self._xba_add_macromolecule_processing_reactions()
        self.parameter_values = self._xba_get_parameter_values(growth_rate)
        self._xba_add_enzyme_concentration_variables(growth_rate)
        self._xba_add_pm_concentration_variables(growth_rate)
        self._xba_add_macromolecule_target_conc_variables(growth_rate)
        self._xba_add_metabolite_target_conc_variables(growth_rate)
        self._xba_add_target_density_variables()
        self._xba_add_flux_targets()
        self._xba_set_dummy_fba_objective()
        # self._xba_unblock_exchange_reactions()
        self.initial_assignments.xba_implementation(self.model)
        self.model.clean(protect_ids)

        # add additional units required for additional parameters
        unit_id = 'mmol_per_l'
        u_dict = {'id': unit_id, 'name': 'millimole per liter',
                  'units': 'kind=mole, exp=1.0, scale=-3, mult=1.0; kind=litre, exp=-1.0, scale=0, mult=1.0'}
        if unit_id not in self.model.parameters:
            self.model.add_unit_def(u_dict)

        # add some parameter values for reference (after model.clean() so they are not removed)
        add_params = {'avg_enz_sat': {'value': self.avg_enz_sat, 'name': 'average enzyme saturation level'},
                      'default_importer_km_value': {'value': DEFAULT_MICHAELIS_CONSTANT, 'name': 'importer Km value',
                                                    'units': 'mmol_per_l'}}
        for pid, p_data in add_params.items():
            self.model.add_parameter(pid, p_data)

        # modify some model attributes and create L3V2 SBML model -
        #  set fbcStrict to False, which is required to supports immediate assignments
        self.model.model_attrs['fbcStrict'] = False
        self.model.model_attrs['id'] += f'_RBA'
        if 'name' in self.model.model_attrs:
            self.model.model_attrs['name'] = f'RBA model of ' + self.model.model_attrs['name']
        self.model.sbml_container['level'] = 3
        self.model.sbml_container['version'] = 2

        print('XbaModel updated with RBA configuration')

    @staticmethod
    def _check_rba_params_functions(rba_params):
        """Check functions used in RBA parameters file.

        Functions used in aggregates of 'compartments', 'targets'
        of 'processes' sheets must also be defined in 'functions'

        :param dict(str, pandas.DataFrame) rba_params:
        :return: success status
        :rtype: bool
        """
        ok_flag = True

        # check defined functions
        defined_functions = set()
        if 'functions' in rba_params:
            defined_functions = set(rba_params['functions'].index)

        # collect functions used in 'compartments', 'targets' sheet
        used_functions = defaultdict(set)
        if 'compartments' in rba_params:
            for cid, row in rba_params['compartments'].iterrows():
                agg = row.get('translation_target_aggregate')
                if type(agg) is str:
                    for fid in [item.strip() for item in agg.split(',')]:
                        used_functions[fid].add(f'compartments: {cid}')
                agg = row.get('density_constraint_aggregate')
                if type(agg) is str:
                    for fid in [item.strip() for item in agg.split(',')]:
                        used_functions[fid].add(f'compartments: {cid}')

        if 'targets' in rba_params:
            for tgid, row in rba_params['targets'].iterrows():
                agg = row.get('target_aggregate')
                if type(agg) is str:
                    for fid in [item.strip() for item in agg.split(',')]:
                        used_functions[fid].add(f'targets: {tgid}')

        if 'processes' in rba_params:
            for pid, row in rba_params['processes'].iterrows():
                agg = row.get('capacity_aggregate')
                if type(agg) is str:
                    for fid in [item.strip() for item in agg.split(',')]:
                        used_functions[fid].add(f'processes: {pid}')

        # check that functions have been defined
        for fid, xids in used_functions.items():
            if fid not in defined_functions:
                print(f'{fid:35} function not defined, but used in {xids}')
                ok_flag = False

        default_fids = {'zero', 'default_spontaneous'}
        for fid in defined_functions:
            if fid not in used_functions and fid not in default_fids:
                print(f'{fid:35s} function defined, but not used')
        return ok_flag

    def _check_rba_params_labels(self, rba_params):
        """Check gene labels used in RBA Parameters file.

        Labels used 'trna2locus' and 'machineries' must exist
        in model UniProt data or NCBI data (for rnas)

        :param dict(str, pandas.DataFrame) rba_params:
        :return: success status
        :rtype: bool
        """
        ok_flag = True
        if 'trna2locus' in rba_params:
            for rna_id, row in rba_params['trna2locus'].iterrows():
                if row['label'] not in self.model.ncbi_data.locus2record:
                    print(f"{row['label']:35s} not found in ncbi data")
                    ok_flag = False

        if 'machineries' in rba_params:
            for pid, row in rba_params['machineries'].iterrows():
                if row.get('set', '') == 'protein':
                    if (row['label'] not in self.model.uniprot_data.locus2uid and
                            row['label'] not in self.model.ncbi_data.locus2protein):
                        print(f"{row['label']:35s} not found in UniProt nor NCBI data")
                        ok_flag = False
                elif row.get('set', '') == 'rna':
                    if row['label'] not in self.model.ncbi_data.locus2record:
                        print(f"{row['label']:35s} not found in ncbi data")
                        ok_flag = False
        return ok_flag

    def _xba_add_macromolecule_species(self):
        """Add macromolecules to XbaModel species (proteins, dna, rna).

        Constraint ID: 'MM_<mmid>

        Macromolecules (except for small macromolecules, i.e. single dna, mrna single components)
        are scaled at a fixed factor of 1000, i.e. µmol instead of mmol (improves LP problem) for large
        macromolecules, i.e. > 10 amino acids weight.
        The scale factor is set on the macromolecule instance.

        MIRIAM annotation and Uniport id is added
        XML annotation with equivalent amino acid weight and scale factor is added

        Note: production / degradation reaction and target concentration variables are scaled accordingly.
        """
        xml_prefix = f'ns_uri={XML_SPECIES_NS}, prefix=rba, token=macromolecule'

        mm_species = {}
        # create protein species and configure scale
        for mm_id, mm in self.proteins.macromolecules.items():
            mm.sid = pf.MM_ + valid_sbml_sid(mm_id)
            mm.scale = 1000.0 if mm.weight > 10 else 1.0
            # for indiviually modeled proteins (UniProt data is collected)
            if mm_id in self.model.locus2uid:
                uid = self.model.locus2uid[mm_id]
                name = f'macromolecule {mm_id} ({uid})'
                miriam_annot = f'bqbiol:is, uniprot/{uid}'
                mw_kda = self.model.proteins[uid].mw / 1000.0
            # for dummy proteins, we calculate molecular weight based on aa sequence
            else:
                mw_kda = protein_mw_from_aa_comp(mm.composition) / 1000.0
                miriam_annot = None
                name = f'macromolecule {mm_id}'
            xml_annot = (f'{xml_prefix}, type={self.proteins.type}, '
                         f'weight_aa={mm.weight}, weight_kDa={mw_kda:.4f}, scale={mm.scale}')
            mm_species[mm.sid] = [name, mm.compartment, False, False, False,
                                  f'meta_{mm.sid}', miriam_annot, xml_annot]

        for mm_id, mm in self.rnas.macromolecules.items():
            mm.sid = pf.MM_ + valid_sbml_sid(mm_id)
            mm.scale = 1000.0 if mm.weight > 10 else 1.0
            mw_kda = rna_mw_from_nt_comp(mm.composition) / 1000.0
            name = f'macromolecule {mm_id}'
            xml_annot = (f'{xml_prefix}, type={self.rnas.type}, '
                         f'weight_aa={mm.weight}, weight_kDa={mw_kda:.4f}, scale={mm.scale}')
            mm_species[mm.sid] = [name, mm.compartment, False, False, False, f'meta_{mm.sid}', None, xml_annot]

        for mm_id, mm in self.dna.macromolecules.items():
            mm.sid = pf.MM_ + valid_sbml_sid(mm_id)
            mm.scale = 1000.0 if mm.weight > 10 else 1.0
            mw_kda = ssdna_mw_from_dnt_comp(mm.composition) / 1000.0
            name = f'macromolecule {mm_id}'
            xml_annot = (f'{xml_prefix}, type={self.dna.type}, '
                         f'weight_aa={mm.weight}, weight_kDa={mw_kda:.4f}, scale={mm.scale}')
            mm_species[mm.sid] = [name, mm.compartment, False, False, False, f'meta_{mm.sid}', None, xml_annot]

        cols = ['name', 'compartment', 'hasOnlySubstanceUnits', 'boundaryCondition', 'constant',
                'meta_id', 'miriamAnnotation', 'xmlAnnotation']
        df_add_species = pd.DataFrame(mm_species.values(), index=list(mm_species), columns=cols)
        print(f"{len(df_add_species):4d} macromoledules to add")
        self.model.add_species(df_add_species)

    def _xba_add_rba_constraints(self):
        """Add RBA constraints to XbaModel instance.

        constrains are added as pseudo species (similar to mass balance constraints of metabolites)
        'C_PMC_<prod_id>: process machine capacities
        'C_EF_<eid>: forward enzyme capacities
        'C_ER_<eid>: reverse enzyme capacities
        'C_D_<cid>: compartment density constraints
        """
        constraints = {}
        for proc_id, proc in self.processes.processes.items():
            if 'capacity' in proc.machinery:
                constr_id = pf.C_PMC_ + valid_sbml_sid(proc_id)
                constraints[constr_id] = [f'capacity {proc_id}', self.cid_mappings['cytoplasm_cid'],
                                          False, False, False]
                proc.constr_id = constr_id

        for eid in sorted(self.enzymes.enzymes):
            e = self.enzymes.enzymes[eid]
            if len(e.mach_reactants) > 0 or len(e.mach_products) > 0:
                eidx = re.sub(r'_enzyme$', '', eid)
                if e.forward_eff != self.parameters.f_name_zero:
                    constr_id = pf.C_EF_ + valid_sbml_sid(eidx)
                    constraints[constr_id] = [f'fwd efficiency {eid}', self.cid_mappings['cytoplasm_cid'],
                                              False, False, False]
                    e.constr_id_fwd = constr_id
                if e.backward_eff != self.parameters.f_name_zero:
                    constr_id = pf.C_ER_ + valid_sbml_sid(eidx)
                    constraints[constr_id] = [f'rev efficiency {eid}', self.cid_mappings['cytoplasm_cid'],
                                              False, False, False]
                    e.constr_id_rev = constr_id

        for cid, d in self.densities.densities.items():
            constr_id = pf.C_D_ + valid_sbml_sid(cid)
            constraints[constr_id] = [f'compartment density for {cid}', cid, False, False, False]
            d.constr_id = constr_id

        cols = ['name', 'compartment', 'hasOnlySubstanceUnits', 'boundaryCondition', 'constant']
        df_add_species = pd.DataFrame(constraints.values(), index=list(constraints), columns=cols)
        print(f"{len(df_add_species):4d} RBA constraints to add")
        self.model.add_species(df_add_species)

    def _xba_couple_reactions(self):
        """Couple reactions to enzyme efficiencies.

        i.e. respective forward enzyme efficiency added as product to enzyme catalyzed reaction
             respective reverse enzyme efficiency added as substrate to reversible enzyme catalyzed reaction

        Note: for reverse reaction enzyme coupling RBA and TRBA reactions have to be
          coupled differently to the enzyme.
            - RBA: C_ER_<ridx> coupled with -1 to reaction rid
            - TFBA: C_ER_<ridx> coupled with +1 to reaction rid_REV

        Enzyme coupling constraint:
            C_EF_<ridx>_enzyme: R_<ridx> - kcat * V_EC_<ridx> ≤ 0
            C_ER_<ridx>_enzyme: -1 R_<ridx> - kcat * V_EC_<ridx> ≤ 0
        """
        modify_attrs = []
        for eid in sorted(list(self.enzymes.enzymes)):
            e = self.enzymes.enzymes[eid]
            if len(e.mach_reactants) > 0 or len(e.mach_products) > 0:
                rid = e.reaction
                if e.forward_eff != self.parameters.f_name_zero:
                    modify_attrs.append([e.reaction, 'reaction', 'product', f'{e.constr_id_fwd}=1.0'])
                if e.backward_eff != self.parameters.f_name_zero:
                    if e.rev_reaction == rid:
                        modify_attrs.append([rid, 'reaction', 'reactant', f'{e.constr_id_rev}=1.0'])
                    else:
                        modify_attrs.append([e.rev_reaction, 'reaction', 'product', f'{e.constr_id_rev}=1.0'])

        cols = ['id', 'component', 'attribute', 'value']
        df_modify_attrs = pd.DataFrame(modify_attrs, columns=cols)
        df_modify_attrs.set_index('id', inplace=True)
        print(f'{len(df_modify_attrs):4d} fwd/rev reactions to couple with enzyme efficiency constraints')
        self.model.modify_attributes(df_modify_attrs, 'reaction')

    def _xba_add_macromolecule_processing_reactions(self):
        """Create production and degradation reactions for macromolecules.

        Here, macromolecule production/degradation reactions are implemented in detail, whereas these
        are lumped in the original RBA formulation, making it difficult to follow the implementation in RBApy.

        Though this will add more variables to the linear problem, the advantages of separating the
        individual cost terms (to better understand the underlying mechanism) could improve
        understanding of the model construction.

        for specified macromolecules we add up production requirements and create a 'R_PROC_<mm_id> reaction
        for specified macromolecules we add up degradation requirements and create a 'R_DEGR_<mm_id> reaction

        processing costs are identified and added to processing constraints 'C_PMC_<pm_id>', actually
        adding respective products with calculated costs as stoic coefficients.

        reaction fluxes are in units of µmol/gDWh (note: metabolic reaction fluxes are in mmol/gDWh)
        """
        # get mapping of macromolecule to related processing reactions for production and degradation:
        mm_productions = defaultdict(list)
        mm_degradations = defaultdict(list)
        for proc_id, proc in self.processes.processes.items():
            for mm_id in proc.productions.get('inputs', []):
                mm_productions[mm_id].append(proc_id)
            for mm_id in proc.degradations.get('inputs', []):
                mm_degradations[mm_id].append(proc_id)

        # determine processing cost for productions / degradations and create processing reactions
        lb_pid_mmol = self.model.fbc_shared_pids[self.model.flux_uid][0.0]
        ub_pid_mmol = self.model.get_fbc_bnd_pid(MAX_MM_PROCESSING_FLUX, self.model.flux_uid,
                                                 'max_processing_flux_mmol')
        lb_pid_umol = self.model.get_fbc_bnd_pid(0.0, 'umol_per_gDWh', 'zero_processing_flux_umol')
        ub_pid_umol = self.model.get_fbc_bnd_pid(MAX_MM_PROCESSING_FLUX, 'umol_per_gDWh', 'max_processing_flux_umol')

        prod_reactions = {}
        degr_reactions = {}
        for mm_id in sorted(self.mmid2mm):
            mm = self.mmid2mm[mm_id]
            # separately check production processes and degradation processes
            for proc_type, mm2maps in {'PROD': mm_productions, 'DEGR': mm_degradations}.items():
                reactants = defaultdict(float)
                products = defaultdict(float)
                if mm2maps.get(mm_id):
                    # if there is any production/degradation process create combined production/degradation fluxes
                    for pm_id in mm2maps.get(mm_id):
                        pm = self.processes.processes[pm_id]
                        pm_map_id = pm.productions['processingMap'] if proc_type == 'PROD' \
                            else pm.degradations['processingMap']
                        pm_map = self.processes.processing_maps[pm_map_id]

                        # constant processing requirements
                        for sid, s_stoic in pm_map.constant_processing.get('reactants', {}).items():
                            reactants[sid] += float(s_stoic)/mm.scale
                        for sid, s_stoic in pm_map.constant_processing.get('products', {}).items():
                            products[sid] += float(s_stoic)/mm.scale

                        # component processing requirements
                        for comp_id, comp_stoic in mm.composition.items():
                            if comp_id in pm_map.component_processings:
                                comp_proc = pm_map.component_processings[comp_id]
                                for sid, s_stoic in comp_proc.get('reactants', {}).items():
                                    reactants[sid] += comp_stoic * float(s_stoic)/mm.scale
                                for sid, s_stoic in comp_proc.get('products', {}).items():
                                    products[sid] += comp_stoic * float(s_stoic)/mm.scale

                        # processing cost, wrt processing machine capacity, if any
                        if pm.constr_id:
                            pm_costs = 0.0
                            for comp_id, comp_stoic in mm.composition.items():
                                if comp_id in pm_map.component_processings:
                                    comp_proc = pm_map.component_processings[comp_id]
                                    pm_costs += comp_stoic * comp_proc.get('cost', 0.0)
                            products[pm.constr_id] = round(pm_costs / mm.scale, 8)

                    if proc_type == 'PROD':
                        rid = pf.R_PROD_ + valid_sbml_sid(mm_id)
                        products[mm.sid] += 1.0
                        if mm.scale == 1.0:
                            prod_reactions[rid] = [f'production reaction for {mm_id} (mmol)', False,
                                                   dict(reactants), dict(products), mm.scale,
                                                   lb_pid_mmol, ub_pid_mmol,
                                                   'RBA_pm_reaction', 'RBA macromolecule processing']
                        else:
                            prod_reactions[rid] = [f'production reaction for {mm_id} (µmol)', False,
                                                   dict(reactants), dict(products), mm.scale,
                                                   lb_pid_umol, ub_pid_umol,
                                                   'RBA_pm_reaction', 'RBA macromolecule processing']
                    else:
                        rid = pf.R_DEGR_ + valid_sbml_sid(mm_id)
                        reactants[mm.sid] += 1.0
                        if mm.scale == 1.0:
                            degr_reactions[rid] = [f'degradation reaction for {mm_id} (mmol)', False,
                                                   dict(reactants), dict(products), mm.scale,
                                                   lb_pid_mmol, ub_pid_mmol,
                                                   'RBA_pm_reaction', 'RBA macromolecule processing']
                        else:
                            degr_reactions[rid] = [f'degradation reaction for {mm_id} (µmol)', False,
                                                   dict(reactants), dict(products), mm.scale,
                                                   lb_pid_umol, ub_pid_umol,
                                                   'RBA_pm_reaction', 'RBA macromolecule processing']

        pm_reactions = prod_reactions | degr_reactions
        cols = ['name', 'reversible', 'reactants', 'products', 'scale',
                'fbcLowerFluxBound', 'fbcUpperFluxBound', 'kind', 'notes']
        df_add_rids = pd.DataFrame(pm_reactions.values(), index=list(pm_reactions), columns=cols)
        print(f'{len(df_add_rids):4d} processing reactions to add')
        self.model.add_reactions(df_add_rids)

    def _xba_get_parameter_values(self, growth_rate):
        """Calculate parameter values based on given growth rate.

        :param float growth_rate: growth rate in h-1 for which to determine values
        :return: parameter values determined based on params
        :rtype: dict(str, float)
        """
        medium_cid = self.cid_mappings['medium_cid']
        mid2sid = {}
        for mid in self.medium.concentrations:
            if f'{mid}_{medium_cid}' in self.model.species:
                mid2sid[mid] = f'{mid}_{medium_cid}'

        params = {mid2sid[mid]: val for mid, val in self.medium.concentrations.items() if val != 0.0}
        params['growth_rate'] = growth_rate
        values = self.parameters.get_values(params)
        print(f"{len(values):4d} parameter values calulated based on {growth_rate:.2f} h-1 growth rate and medium")
        return values

    def _couple_weights(self, srefs):
        """Calculate enzymes/process machines weight and couple to density constraints.

        :param dict(str, float) srefs: species reference with reactants and stoic coefficient of composition
        :return: density constraints affected by weight
        :rtype: dict(str, float)
        """
        weights = defaultdict(float)
        for comp_id, stoic in srefs.items():
            if comp_id in self.mmid2mm:
                mm = self.mmid2mm[comp_id]
                weight = stoic * mm.weight/mm.scale
                weights[mm.compartment] += weight
                if 'total_density_cid' in self.cid_mappings:
                    weights[self.cid_mappings['total_density_cid']] += weight

        weight_coupling = defaultdict(float)
        for cid, weight in weights.items():
            if cid in self.densities.densities:
                weight_coupling[self.densities.densities[cid].constr_id] = weight

        return dict(weight_coupling)

    def _get_dilution_srefs(self, var_id, growth_rate, composition_srefs):
        """Determine reactant or product species references for concentration variables.

        These variables are in units of µmol/gDW. Coefficients need to be scaled

        RBA macromolecule ids are converted to respective XbaModel species id
        Reactants/products are based on stoichiometry of machinery composition
        Reactants/products are diluted as per growth rate
        Reactants/products are scaled to match units of concentration variable (µmol_per_gDW)
        As stoichiometry is based on growth rate, data for initial assignment is collected
        as well as species reference ids

        SBML Validation:
          When the value of the attribute variable in an InitialAssignment object refers to a SpeciesReference object,
          the unit of measurement associated with the InitialAssignment's math expression should be consistent with
          the unit of stoichiometry, that is, dimensionless.

        :param str var_id: id of variable, e.g. 'V_EC_<eid>', required for sref id construction
        :param float growth_rate: growth rate in h-1, dilute molecule accordingly
        :param dict(str, float) composition_srefs: reactants or product species refs
        :return: srefs with stoichiometry
        :rtype: dict(str, float)
        """
        dilution_srefs = {}
        for m_sid, m_stoic in composition_srefs.items():
            if m_sid in self.mmid2mm:
                sid = self.mmid2mm[m_sid].sid
                scale = self.mmid2mm[m_sid].scale
            else:
                sid = m_sid
                scale = 1.0
            if scale == 1000.0:
                dilution_srefs[sid] = growth_rate * m_stoic
                self.initial_assignments.add_sref_ia(var_id, sid, math_var=f'growth_rate * {m_stoic} hour')
            else:
                dilution_srefs[sid] = growth_rate * m_stoic / 1000.0
                self.initial_assignments.add_sref_ia(var_id, sid,
                                                     math_var=f'growth_rate * {m_stoic/ 1000.0} hour')
        return dilution_srefs

    def _xba_add_enzyme_concentration_variables(self, growth_rate):
        """Add enzyme concentration variables to the XbaModel instance.

        Enzyme concentration variables: V_EC_<ridx>
        Couple variables to macromolecule mass balances.
        Couple these variables to enzyme efficiency and process machinery
        capacity constraints.
        Couple the variables to compartment density constraints.

        Enzyme concentrations are in µmol/gDW

        :param float growth_rate: growth rate (h-1) required for macromolecule dilution
        """
        xml_prefix = f'ns_uri={XML_SPECIES_NS}, prefix=rba, token=macromolecule'

        lb_pid = self.model.get_fbc_bnd_pid(0.0, 'umol_per_gDW', 'zero_enzyme_umol_conc')
        ub_pid = self.model.get_fbc_bnd_pid(MAX_ENZ_CONC, 'umol_per_gDW', 'max_enzyme_umol_conc')
        scale = 1000.0

        conc_vars = {}
        for eid, e in self.enzymes.enzymes.items():
            if len(e.mach_reactants) > 0 or len(e.mach_products) > 0:
                eidx = re.sub(r'_enzyme$', '', valid_sbml_sid(eid))
                var_id = pf.V_EC_ + eidx
                name = f'{eid} concentration (µmol)'

                # Determine reactants/products and sref ids for enzyme concentration variable
                reactants = self._get_dilution_srefs(var_id, growth_rate, e.mach_reactants)
                products = self._get_dilution_srefs(var_id, growth_rate, e.mach_products)

                # get gene product association from proteins used in machinery
                gps = [self.model.locus2gp[locus] for locus in e.mach_reactants if locus in self.model.locus2gp]
                gpa = 'assoc=(' + ' and '.join(gps) + ')' if len(gps) > 0 else None

                # enzyme efficiency constraints coupling
                if e.constr_id_fwd:
                    reactants[e.constr_id_fwd] = self.parameter_values[e.forward_eff] / scale
                    self.initial_assignments.add_sref_ia(var_id, e.constr_id_fwd, rba_pid=e.forward_eff,
                                                         math_const=f'{1.0/scale} dimensionless')
                if e.constr_id_rev:
                    reactants[e.constr_id_rev] = self.parameter_values[e.backward_eff] / scale
                    self.initial_assignments.add_sref_ia(var_id, e.constr_id_rev, rba_pid=e.backward_eff,
                                                         math_const=f'{1.0/scale} dimensionless')

                # couple weights of enzyme composition to density constraints
                products.update(self._couple_weights(e.mach_reactants))

                # determine enzyme molecular weight
                enz_mw_kda = 0.0
                for comp_id, stoic in e.mach_reactants.items():
                    if comp_id in self.proteins.macromolecules:
                        if comp_id in self.model.locus2uid:
                            uid = self.model.locus2uid[comp_id]
                            mw_kda = self.model.proteins[uid].mw / 1000.0
                            enz_mw_kda += stoic * mw_kda
                xml_annot = f'{xml_prefix}, weight_kDa={enz_mw_kda:.4f}, scale={scale}'

                conc_vars[var_id] = [name, False, reactants, products, scale, gpa, lb_pid, ub_pid,
                                     'RBA_enz_conc', xml_annot]

        cols = ['name', 'reversible', 'reactants', 'products', 'scale', 'fbcGeneProdAssoc',
                'fbcLowerFluxBound', 'fbcUpperFluxBound', 'kind', 'xmlAnnotation']
        df_add_rids = pd.DataFrame(conc_vars.values(), index=list(conc_vars), columns=cols)
        print(f'{len(df_add_rids):4d} enzyme concentration variables to add')
        self.model.add_reactions(df_add_rids)

    def _xba_add_pm_concentration_variables(self, growth_rate):
        """Add process machine concentration variables to the XbaModel instance.

        see _xba_add_enzyme_concentration_variables
        Process machinery concentration variables: V_PMC_<pm_id>
        Couple variables to macromolecule mass balances.
        Couple these variables to enzyme efficiency and process machinery
        capacity constraints.
        Couple the variables to compartment density constraints.

        Process machinery concentrations are in µmol/gDW

        :param float growth_rate: growth rate (h-1) required for macromolecule dilution
        """
        xml_prefix = f'ns_uri={XML_SPECIES_NS}, prefix=rba, token=macromolecule'

        lb_pid = self.model.get_fbc_bnd_pid(0.0, 'umol_per_gDW', 'zero_enzyme_umol_conc')
        ub_pid = self.model.get_fbc_bnd_pid(MAX_ENZ_CONC, 'umol_per_gDW', 'max_enzyme_umol_conc')
        scale = 1000.0

        conc_vars = {}
        for pm_id, pm in self.processes.processes.items():
            if 'capacity' in pm.machinery:
                var_id = pf.V_PMC_ + valid_sbml_sid(pm_id)
                name = f'{pm_id} concentration'

                # Determine reactants/products and sref ids for PM concentration variable
                reactants = self._get_dilution_srefs(var_id, growth_rate, pm.machinery['reactants'])
                products = self._get_dilution_srefs(var_id, growth_rate, pm.machinery['products'])

                # get gene product association
                gps = [self.model.locus2gp[locus] for locus in pm.machinery['reactants']
                       if locus in self.model.locus2gp]
                gpa = 'assoc=(' + ' and '.join(gps) + ')' if len(gps) > 0 else None

                # process machinery capacity coupling
                reactants[pm.constr_id] = self.parameter_values[pm.machinery['capacity'].value] / scale
                self.initial_assignments.add_sref_ia(var_id, pm.constr_id, rba_pid=pm.machinery['capacity'].value,
                                                     math_const=f'{1.0 / scale} dimensionless')

                # couple weights of process machine composition to density constraints
                products.update(self._couple_weights(pm.machinery['reactants']))

                # determine molecular weight of process machine, consisting of proteins and/or RNA
                pm_mw_kda = 0.0
                for comp_id, stoic in pm.machinery['reactants'].items():
                    if comp_id in self.proteins.macromolecules:
                        if comp_id in self.model.locus2uid:
                            uid = self.model.locus2uid[comp_id]
                            mw_kda = self.model.proteins[uid].mw / 1000.0
                            pm_mw_kda += stoic * mw_kda
                    elif comp_id in self.rnas.macromolecules:
                        mm = self.rnas.macromolecules[comp_id]
                        mw_kda = rna_mw_from_nt_comp(mm.composition) / 1000.0
                        pm_mw_kda += stoic * mw_kda
                xml_annot = f'{xml_prefix}, weight_kDa={pm_mw_kda:.4f}, scale={scale}'

                conc_vars[var_id] = [name, False, reactants, products, scale, gpa, lb_pid, ub_pid,
                                     'RBA_pm_conc', xml_annot]

        cols = ['name', 'reversible', 'reactants', 'products', 'scale', 'fbcGeneProdAssoc',
                'fbcLowerFluxBound', 'fbcUpperFluxBound', 'kind', 'xmlAnnotation']
        df_add_rids = pd.DataFrame(conc_vars.values(), index=list(conc_vars), columns=cols)
        print(f'{len(df_add_rids):4d} process machine concentration variables to add')
        self.model.add_reactions(df_add_rids)

    def _xba_add_macromolecule_target_conc_variables(self, growth_rate):
        """Add variables for macromolecule target concentrations to the XbaModel instance.

        V_TMMC_<mid>: For each target concentration of a macromolecule a separate variable is introduced.
        Units: µmol/gDW or mmol/gDW
        This variable is fixed (lower/upper bound) to target concentration in respective unit (this can be
        a function of growth rate). Note: targets concentrations are defined in mmol/gDW and need to be rescaled
        Reactant coefficient for macromolecule is 1 times growth rate (i.e. macromolecular dilution),
        Product coefficient / Weight coupling is macromolecular weight divided by scale (weight in mmol AA/gDW)
        Variable bounds need to be updated during bisection optimization.

        :param float growth_rate: growth rate (h-1) required for macromolecule dilution
        """
        conc_targets = {}
        for tg_id, tg in self.targets.target_groups.items():
            for tid, t in tg.concentrations.items():
                if tid in self.mmid2mm:
                    var_id = pf.V_TMMC_ + valid_sbml_sid(tid)
                    mm = self.mmid2mm[tid]

                    # macromolecule dilution with growth rate
                    reactants = {mm.sid: growth_rate}
                    self.initial_assignments.add_sref_ia(var_id, mm.sid, math_var='growth_rate * 1 hour')

                    # add compartment density constraint
                    weight_coupling = {}
                    cids = [mm.compartment]
                    if 'total_density_cid' in self.cid_mappings:
                        cids.append(self.cid_mappings['total_density_cid'])
                    for cid in cids:
                        if cid in self.densities.densities:
                            weight_coupling[self.densities.densities[cid].constr_id] = mm.weight / mm.scale
                    products = weight_coupling

                    # get gene product association
                    gpid = self.model.locus2gp.get(tid)
                    gpa = f'assoc={gpid}' if gpid else None

                    # target concentration implemented as fixed variable bound
                    units_id = 'mmol_per_gDW' if mm.scale == 1.0 else 'umol_per_gDW'
                    var_bnd_val = self.parameter_values[t.value] * mm.scale
                    var_bnd_pid = self.model.get_fbc_bnd_pid(var_bnd_val, units_id, f'target_conc_{tid}', reuse=False)
                    conc_targets[var_id] = [f'{tid} concentration target', False, reactants, products,
                                            mm.scale, gpa, var_bnd_pid, var_bnd_pid, 'RBA_mm_conc']
                    self.initial_assignments.add_var_bnd_ia(var_bnd_pid, rba_pid=t.value,
                                                            math_const=f'{mm.scale} {units_id}')

        cols = ['name', 'reversible', 'reactants', 'products', 'scale', 'fbcGeneProdAssoc',
                'fbcLowerFluxBound', 'fbcUpperFluxBound', 'kind']
        df_add_rids = pd.DataFrame(conc_targets.values(), index=list(conc_targets), columns=cols)
        print(f'{len(df_add_rids):4d} macromolecule concentration target variables to add')
        self.model.add_reactions(df_add_rids)

    def _xba_add_metabolite_target_conc_variables(self, growth_rate):
        """Add variables for metabolite (small molecules) target concentrations to the XbaModel instance.

        V_TSMC: Target concentrations for small molecules are collected in a single variable
        Unit: µmol/gDW
        The variable is fixed at the growth rate (lower/upper bound)
        Reactant coefficients are the target concentrations (these can be functions of growth rate by itself)
        Growth rate dependent coefficients and variable bounds need to be updated during bisection optimization.
           NOTE: THIS will be changed to target = 1 and dilution terms for molecules

        :param float growth_rate: growth rate (h-1) required for macromolecule dilution
        """
        # target concentration variables are constants, i.e. fixed at value 1
        var_id = pf.V_TSMC
        one_umol_pid = self.model.get_fbc_bnd_pid(1.0e-3, 'mmol_per_gDW', 'one_umol_per_gDW', reuse=False)
        scale = 1000.0

        conc_targets = {}
        reactants_tsmc = {}
        for tg_id, tg in self.targets.target_groups.items():
            for sid, t in tg.concentrations.items():
                sbml_sid = valid_sbml_sid(sid)
                if sid not in self.mmid2mm:
                    assert sid in self.model.species, f'targets: metabolite {sid} is not included in model species'
                    reactants_tsmc[sbml_sid] = growth_rate * self.parameter_values[t.value] * scale
                    self.initial_assignments.add_sref_ia(var_id, sbml_sid, rba_pid=t.value,
                                                         math_var=f'growth_rate * {scale} hour')
        conc_targets[var_id] = [f'small molecules concentration targets (mmol)', False, reactants_tsmc, {},
                                scale, one_umol_pid, one_umol_pid, 'RBA_sm_conc']

        cols = ['name', 'reversible', 'reactants', 'products', 'scale',
                'fbcLowerFluxBound', 'fbcUpperFluxBound', 'kind']
        df_add_rids = pd.DataFrame(conc_targets.values(), index=list(conc_targets), columns=cols)
        print(f'{len(df_add_rids):4d} metabolites concentration target variable to add')
        self.model.add_reactions(df_add_rids)

    def _xba_add_target_density_variables(self):
        """Add variables for target compartment density and slack variable to XbaModel.

        V_TCD: Target density concentrations
        Units: mmol AA/gDW
        The variable is fixed at 1.
        Reactant coefficients are the target densities for the various compartments
        Target concentrations that depend on growth rate need to be updated during bisection optimization

        V_SLACK_<cid>: positive slack on compartment density constraints
        Units: mmol AA/gDW
        Variable is bounded by upper value of MAX_DENSITY_SLACK (here 10 mmol AA/gDW)
        Implemented, so we can report on maximum compartment capacity utilization.
        """
        # add compartment density targets
        var_id = pf.V_TCD
        one_mmol_pid = self.model.get_fbc_bnd_pid(1.0, 'mmol_per_gDW', 'one_mmol_aa_per_gDW', reuse=False)
        density_vars = {}
        reactants = {}
        reactants_vars = {}
        for cid, d in self.densities.densities.items():
            reactants[d.constr_id] = self.parameter_values[d.target_value.upper_bound]
            self.initial_assignments.add_sref_ia(var_id, d.constr_id, rba_pid=d.target_value.upper_bound)

            fid = d.target_value.upper_bound
            if not (fid in self.parameters.functions and self.parameters.functions[fid].type == 'constant'):
                reactants_vars[d.constr_id] = fid
        density_vars[var_id] = [f'max compartment density target (mmol AA)', False, reactants, {},
                                one_mmol_pid, one_mmol_pid, 'RBA_max_density']

        # add slack variables for density constraints
        lb_pid = self.model.get_fbc_bnd_pid(0.0, 'mmol_per_gDW', 'density_slack_min', reuse=False)
        ub_pid = self.model.get_fbc_bnd_pid(MAX_DENSITY_SLACK, 'mmol_per_gDW', 'density_slack_max', reuse=False)
        for cid, d in self.densities.densities.items():
            products = {d.constr_id: 1.0}
            density_vars[pf.V_SLACK_ + cid] = [f'Positive Slack on {cid} density', False, {}, products,
                                               lb_pid, ub_pid, 'slack_variable']

        cols = ['name', 'reversible', 'reactants', 'products',
                'fbcLowerFluxBound', 'fbcUpperFluxBound', 'kind']
        df_add_rids = pd.DataFrame(density_vars.values(), index=list(density_vars), columns=cols)
        print(f'{len(df_add_rids):4d} density target and slack variables to add')
        self.model.add_reactions(df_add_rids)

    def _xba_add_flux_targets(self):
        """Add RBA reaction/production/degradation flux targets to the XbaModel instance.

        Such targets are implemented as reaction/variable bounds (fbc_bounds)

        reactionFlux targets are applied to model reactions
        productionFLux and degradatationFlux are applied on macromolecule production/degradation reactions.
        Note: a productionFlux target is implemented as a lower bound.

        reaction/variable bound parameters are created with 'correct' units.

        We also collect the flux bounds depending on growth rate, to be applied during optimization.
        """
        non_const_fids = defaultdict(dict)

        modify_attrs = []
        for tg_id, tg in self.targets.target_groups.items():

            flux_targets = {'': tg.reaction_fluxes, 'R_PROD_': tg.production_fluxes, 'R_DEGR_': tg.degradation_fluxes}
            for rid_prefix, targets in flux_targets.items():
                for tid, t in targets.items():
                    rid = f'{rid_prefix}{tid}'
                    r = self.model.reactions[rid]
                    scale = r.scale
                    ia_units = 'umol_per_gDWh' if scale == 1000.0 else 'mmol_per_gDW_per_hr'
                    unit_id = self.model.parameters[r.fbc_lower_bound].units

                    if t.value:
                        var_bnd_val = self.parameter_values[t.value] * scale
                        var_bnd_pid = self.model.get_fbc_bnd_pid(var_bnd_val, unit_id, f'{rid}_rba_bnd', reuse=False)
                        modify_attrs.append([rid, 'reaction', 'fbc_lower_bound',    var_bnd_pid])
                        self.initial_assignments.add_var_bnd_ia(var_bnd_pid, rba_pid=t.value,
                                                                math_const=f'{scale} {ia_units}')
                        fid = t.value
                        if self.parameters.functions[fid].type != 'constant':
                            non_const_fids[rid]['lb'] = fid

                        if rid_prefix != 'R_PROD_':
                            modify_attrs.append([rid, 'reaction', 'fbc_upper_bound', var_bnd_pid])
                            if self.parameters.functions[fid].type != 'constant':
                                non_const_fids[rid]['ub'] = fid
                    else:
                        if t.lower_bound:
                            var_bnd_val = self.parameter_values[t.lower_bound] * scale
                            var_bnd_pid = self.model.get_fbc_bnd_pid(var_bnd_val, unit_id, f'{rid}_rba_lb', reuse=False)
                            modify_attrs.append([rid, 'reaction', 'fbc_lower_bound', var_bnd_pid])
                            self.initial_assignments.add_var_bnd_ia(var_bnd_pid, rba_pid=t.lower_bound,
                                                                    math_const=f'{scale} {ia_units}')
                            fid = t.lower_bound
                            if self.parameters.functions[fid].type != 'constant':
                                non_const_fids[rid]['lb'] = fid

                        if t.upper_bound:
                            var_bnd_val = self.parameter_values[t.upper_bound] * scale
                            var_bnd_pid = self.model.get_fbc_bnd_pid(var_bnd_val, unit_id, f'{rid}_rba_ub', reuse=False)
                            modify_attrs.append([rid, 'reaction', 'fbc_upper_bound', var_bnd_pid])
                            self.initial_assignments.add_var_bnd_ia(var_bnd_pid, rba_pid=t.upper_bound,
                                                                    math_const=f'{scale} {ia_units}')
                            fid = t.upper_bound
                            if self.parameters.functions[fid].type != 'constant':
                                non_const_fids[rid]['ub'] = fid

        df_modify_attrs = pd.DataFrame(modify_attrs, columns=['id', 'component', 'attribute', 'value'])
        df_modify_attrs.set_index('id', inplace=True)
        print(f"{len(df_modify_attrs):4d} Flux/variable bounds need to be updated.")
        self.model.modify_attributes(df_modify_attrs, 'reaction')

    def _xba_set_dummy_fba_objective(self):
        """Add optimization objective to the XbaModel instance.

        RBA is an LP feasibility problem where, where coefficients and bounds are reconfigured
        depending on a selected growth rate.
        The optimization algorithm of RBA selects for the highest growth rate, which still
        results in a feasible problem. FBA objective is not relevant, though
        in the contexts of FBA optimization (here, just feasibility checking) an (arbitrary)
        FBA objective is defined, so problem can be solved using LP solvers
        """
        # remove any existing objectives
        obj_ids = [obj_id for obj_id in self.model.objectives]
        self.model.del_components('objectives', obj_ids)

        # Configure a dummy FBA objective:
        objectives_config = {'rba_obj': {'type': 'maximize', 'active': True, 'coefficients': {pf.V_TSMC: 1.0}}}
        self.model.add_objectives(objectives_config)
        print(f'Dummy FBA objective configured: maximize {pf.V_TSMC}')

    def to_excel(self, fname):
        """Export RBA model in RBApy format to Microsoft Excel spreadsheet

        :meta private:
        :param str fname: file name of spreadsheet
        :return: success status
        :rtype: bool
        """
        m_dict = {}
        for component in components:
            m_dict |= getattr(self, component).to_df()

        # add target value information strings
        for idx, row in m_dict['enzymes'].iterrows():
            m_dict['enzymes'].at[idx, 'fwd_eff_info'] = self.parameters.get_value_info(row['forwardEfficiency'])
            m_dict['enzymes'].at[idx, 'bwd_eff_info'] = self.parameters.get_value_info(row['backwardEfficiency'])
        for idx, row in m_dict['processes'].iterrows():
            if '=' in row['machineryCapacity']:
                first_value = row['machineryCapacity'].split('=')[1]
                m_dict['processes'].at[idx, 'capacity_info'] = self.parameters.get_value_info(first_value)
        for idx, row in m_dict['densities'].iterrows():
            first_value = row['targetValue'].split('=')[1]
            m_dict['densities'].at[idx, 'value_info'] = self.parameters.get_value_info(first_value)
        # we temporarily create a unique index for table update
        m_dict['targets'].reset_index(inplace=True)
        for idx, row in m_dict['targets'].iterrows():
            first_value = row['targetValue'].split('=')[1]
            m_dict['targets'].at[idx, 'value_info'] = self.parameters.get_value_info(first_value)
        m_dict['targets'].set_index('targetGroup', inplace=True)

        with pd.ExcelWriter(fname) as writer:
            for name, df in m_dict.items():
                df.to_excel(writer, sheet_name=name)
        print(f'RBA model exported to {fname}')
        return True

    def check_unused(self):

        molecules = (set(self.metabolism.species) | set(self.dna.macromolecules) |
                     set(self.rnas.macromolecules) | set(self.proteins.macromolecules))
        parameters = set(self.parameters.functions) | set(self.parameters.aggregates)

        ref_parameters = set()
        ref_parameters |= self.processes.ref_parameters()
        ref_parameters |= self.densities.ref_parameters()
        ref_parameters |= self.targets.ref_parameters()
        ref_parameters |= self.enzymes.ref_parameters()
        functions_used_in_aggs = self.parameters.ref_functions(ref_parameters)
        ref_parameters |= functions_used_in_aggs
        #ref_parameters |= self.parameters.ref_functions(ref_parameters)

        ref_molecules = set()
        ref_molecules |= self.metabolism.ref_molecules()
        ref_molecules |= self.processes.ref_molecules()
        ref_molecules |= self.enzymes.ref_molecules()
        ref_molecules |= self.targets.ref_molecules()

        unused_parameters = parameters.difference(ref_parameters)
        unused_molecules = molecules.difference(ref_molecules)
        unused = 0
        if len(unused_parameters) > 0:
            print(f'{len(unused_parameters)} unused parameters:', unused_parameters)
            unused += len(unused_parameters)
        if len(unused_molecules) > 0:
            print(f'{len(unused_molecules)} unused molecules:', unused_molecules)
            unused += len(unused_molecules)
        if unused == 0:
            print('no unused parameters/molecules')

    def validate(self, validate_sbml=True):
        """Validate compliance with SBML standards and completeness wrt RBA.

        The validation with SBML standards can be suppressed using the optional
        parameter `validate_sbml`.

        Validation against SBML standards is an optional task taking time.
        Validation could be skipped once model configuration is stable.

        Information on non-compliance is printed. Details are written to `tmp/tmp.txt`.
        In case of an unsuccessful validation, it is recommended to review `tmp/tmp.txt` and
        improve on the model configuration.

        Example: Ensure completeness of RbaModel and ensure compliance with SBML standards prior to export.

        .. code-block:: python

            if rba_model.validate():
                rba_model.export('iML1515_RBA.xml')

        :param bool validate_sbml: validate SBML compliance (default: True)
        :return: success status
        :rtype: bool
        """
        component_ids = {'species': set(self.metabolism.species),
                         'dna': set(self.dna.macromolecules),
                         'rna': set(self.rnas.macromolecules),
                         'protein': set(self.proteins.macromolecules),
                         'functions': set(self.parameters.functions),
                         'aggregates': set(self.parameters.aggregates)}
        valid = True
        _components = {'parameters', 'metabolism', 'processes', 'enzymes', 'densities', 'targets'}
        for component in _components:
            valid = valid and getattr(self, component).validate(component_ids)
        print(f'RBA model valid status: {valid}')

        if valid is True and validate_sbml is True:
            print(f'checking SBML compliance ...')
            return self.model.validate()
        else:
            return valid

    def export(self, fname, export_format='SBML'):
        """Export RbaModel instance to SBML encoded file, RBApy encoded format (directory) or spreadsheet (.xlsx).

        Setting the optional parameter `export_format` to `RBApy` and providing a directory name in
        `fname` exports the file in proprietary RBApy 1.0 format. The SBML encoded model
        may include additional features and functionality.

        The spreadsheet (.xlsx) is helpful to inspect model configuration for of SBML and RBApy formats.

        Example: Export RbaModel instance to SBML encoded file and to spreadsheet.

        .. code-block:: python

            xba_model = XbaModel('iML1515.xml')
            xba_model.configure('xba_parameters.xlsx')

            rba_model = RbaModel(xba_model)
            rba_model.configure('rba_parameters.xlsx')
            if rba_model.validate():
                rba_model.export('iML1515_RBA.xml')
                rba_model.export('iML1515_RBA.xlsx')

        :param str fname: filename with extension '.xml' or '.xlsx', or a directory name for RBApy format
        :param str export_format: export format: 'SBML' or 'RBApy' (default: 'SBML')
        :return: success status
        :rtype: bool
        """
        assert export_format in {'SBML', 'RBApy'}, 'argument format must be "SBML" or "RBApy")'
        if export_format == 'SBML':
            return self.model.export(fname)
        elif export_format == 'RBApy':
            if fname.endswith('.xlsx'):
                return self.to_excel(fname)
            else:
                # export RBA model in RBA proprietary format
                if not os.path.exists(fname):
                    os.makedirs(fname)
                    print(f'RBA directory {fname} created')
                for component in components:
                    getattr(self, component).export_xml(fname)
                print(f'RBA model exported to: {fname}')
        return True



"""Implementation of NcbiData class.

Extract nucleotide information from NCBI using E-utils EFetch.

NCBI's Disclaimer and Copyright notice
(https://www.ncbi.nlm.nih.gov/About/disclaimer.html).

also check: NCBI Entrez Programming Utilities Help

Peter Schubert, CCB, HHU Duesseldorf, January 2023
"""

from collections import defaultdict

from .ncbi_chromosome import NcbiChromosome


class NcbiData:
    """Access to data from NCBI nucleotide online resource.

    Resource balance constraint models require access to genome data during model construction.
    Genome data can be referenced by GeneBank or RefSeq accession identifiers.
    Select genome data sets that can be mapped to the gene identifiers used in the model under construction.
    As model genes may be located on different chromosomes, access to several chromosomes is supported.

    Use configuration data in the XBA configuration file, sheet `general`, to set `chromosome2accids`,
    which maps arbitrary chromosome ids to accession ids, and `organism_dir`, were downloaded data
    is stored locally. Delete locally stored NCBI data to enforce a retrieval from the online database.

    Example: Access chromosome data for E. coli K-12 MG1655 strain (accession id: U00096.3).

    .. code-block:: python

        from f2xba.ncbi.ncbi_data import NcbiData

        ncbi_data = NcbiData({'chromosome':'U00096.3'}, 'data_refs/ncbi')

        gene = 'b0928'
        ncbi_data.locus2record[gene].__dict__

    :param dict(str, str) chromosome2accids: Map chromosome ids to accession ids
    :param str organism_dir: directory where NCBI exports are stored
    """

    def __init__(self, chromosome2accids, organism_dir):
        """Instantiate NcbiData with genome data retrieved from NCBI nucleotide online resource.

        Download NCBI nucleotide information for given accession ids.
        Use stored file, if found in organism_dir.

        :param dict(str, str) chromosome2accids: Map chromosome ids to accession ids
        :param str organism_dir: directory where NCBI exports are stored
        """
        self.chromosomes = {}
        """Chromosome related information."""

        for chrom_id, accession_id in chromosome2accids.items():
            self.chromosomes[chrom_id] = NcbiChromosome(chrom_id, accession_id, organism_dir)

        # mapping of NCBI record loci to feature records and proteins across chromosomes
        self.locus2record = {}
        """Map gene identifier to NCBI feature record."""

        self.locus2protein = {}
        """Map gene identifier to NCBI protein sequence information."""

        for chrom_id, chrom in self.chromosomes.items():
            self.locus2record.update(chrom.mrnas)
            self.locus2record.update(chrom.rrnas)
            self.locus2record.update(chrom.trnas)
            self.locus2protein.update(chrom.proteins)

        # mapping of gene product label to NCBI locus (including NCBI old_locus_tag)
        self.label2locus = {}
        """Map gene label to NCBI locus identifiers."""

        self.update_label2locus()

    def update_label2locus(self):
        self.label2locus = {}
        for locus, record in self.locus2record.items():
            self.label2locus[locus] = locus
            if hasattr(record, 'old_locus') and record.old_locus is not None:
                self.label2locus[record.old_locus] = locus

    def modify_attributes(self, df_modify_attrs):
        """modify attribute values of NCBI feature records

        e.g. update 'locus' or 'old_locus' attributes to improve mapping with model loci

        :param pandas.DataFrame df_modify_attrs: table with 'attribute', 'value' columns and index set to gene locus
        """
        for locus, row in df_modify_attrs.iterrows():
            if locus in self.locus2record:
                record = self.locus2record[locus]
                record.modify_attribute(row['attribute'], row['value'])
            else:
                print(f'{locus} not found in NCBI data export')
        self.update_label2locus()

    def get_gc_content(self, chromosome_id=None):
        """Retrieve GC content across all or one specific chromosome

        :param str chromosome_id: (optional) specific chromosome id
        :return: GC content
        :rtype: float
        """
        if chromosome_id is not None:
            chrom_ids = [chromosome_id]
        else:
            chrom_ids = self.chromosomes.keys()

        total_nts = 0
        total_gc = 0
        for chrom_id in chrom_ids:
            chromosome = self.chromosomes[chrom_id]
            total_nts += sum(chromosome.nt_composition.values())
            total_gc += chromosome.nt_composition['G'] + chromosome.nt_composition['C']
        return total_gc / total_nts

    def get_mrna_avg_composition(self, chromosome_id=None):
        """Retrieve average mRNA composition across all or a chromosome

        :param str chromosome_id: (optional) specific chromosome id
        :return: relative mRNA nucleotide composition
        :rtype: dict(str,float)
        """
        chrom_ids = [chromosome_id] if chromosome_id is not None else self.chromosomes.keys()

        nt_comp = defaultdict(int)
        for chrom_id in chrom_ids:
            chrom = self.chromosomes[chrom_id]
            for locus, feature in chrom.mrnas.items():
                for nt, count in feature.spliced_nt_composition.items():
                    nt_comp[nt] += count
        total = sum(nt_comp.values())
        return {nt: count / total for nt, count in nt_comp.items()}



"""Implementation of TdReactant class.

Biochemical reactions deal with reactants, e.g. ATP
Reactants may be composed of several 'chemical' species, e.g. ATP4-, HATP3-, H2ATP2-
TD reactant has one-to-one relation with a model species
TD reactant is also linked to a single TD species with thermodynamics data

holding thermodynamic data for a TD reactant, like ATP
based on pyTFA thermo data

Peter Schubert, HHU Duesseldorf, Octobert 2023
"""

import numpy as np
from f2xba.utils.calc_mw import extract_atoms
from .td_constants import CPD_PROTON, RT, DEBYE_HUCKEL_A, DEBYE_HUCKEL_B, MAX_PH

class TdReactant:

    def __init__(self, s, td_m, td_c):
        """Instantiate thermodynamic reactant data.

        contains reactant TD data for reactant at specific pH and ionic strength
        based on pyTFA thermo data file

        :param s: corresponding model species
        :type s: :class:`SbmlSpecies`
        :param td_m: corresponding TD metabolite
        :type td_m: :class:`TdMetabolite`
        :param td_c: corresponding TD compartment
        :type td_c: :class:`TdCompartment`
        """
        self.id = s.id
        self.s = s
        self.td_m = td_m
        self.td_metabolite = self.td_m.id
        self.td_c = td_c
        self.nh_species = {atom: count for atom, count in extract_atoms(s.formula)}.get('H', 0)
        self.dfg0_tr = 0.0
        self.avg_h_atoms_tr = 0.0
        self.avg_charge_tr = 0.0
        self.mole_fractions_rev = []

    def get_std_pka_level(self):
        """Determine pKa level (deprotonation state) of TD metabolite in standard state.

        Given the electrical charge in standard conditions of the TD metabolite,
        determine the number of deprotonation steps (pKa level), upto MAX_PH
        Using structural cues, we determine any positive charge of the least protonated state.
        We ensure that pKa level does not exceed the length of the pKas provided.

        - we assume an el. charge of 0 for the most protonated state `charge_mp`
        - however, if substructures (cues) of the metabolite carry positive charges, we add those
           charges to charge_mp, considering the stoichiometric coefficient
        Note: in the TD database there are 13 cues with a charge of +1: H, 11 nitrogen containing compounds,
        and WSWW, 1 cue (Fe2) with a charge of +2 and 1 with a charge of +3 (Fe3)

        Examples:
        - pKa level of 0 means, there was no deprotonations required for the TD metabolite
        - pKa level of 2 means, there were two deprotonations, i.e. first two
          pKa value in the sorted list of pKas were used

        :return: pKa level, i.e. number of deprotonations for standard state
        :rtype int
        """
        # ensure pka_level between 0 and number of pKas < MAX_PH
        pka_level = max(0, self.td_m.charge_mp - self.td_m.charge_std)
        pka_level = min(pka_level, sum([1 for pka in self.td_m.pkas if pka < MAX_PH]))

        return int(pka_level)

    def get_dfg0_at_charge(self, target_charge):
        """Determine ∆fG˚ for TD metabolite a given el. charge level

        At zero ionic strength deprotonate or protonate to targe charge level using
        pKa values.
        Process may be incomplete if pKa values or H-atoms get exhausted.

        :param int target_charge: target electrical charge value
        :return: dfg0_target, charge_target and nh_target
        :rtype: float, int, int
        """
        std_pka_level = int(self.get_std_pka_level())

        dfg0_target = self.td_m.dfg0
        charge_target = self.td_m.charge_std
        nh_target = self.td_m.nh_std

        if target_charge < self.td_m.charge_std:
            deprot_steps = self.td_m.charge_std - target_charge
            if deprot_steps > self.td_m.nh_std:
                print(f'{self.id}: incomplete deprotonation due to insufficient H atoms.')
                deprot_steps = self.td_m.nh_std
            deprot_pkas = self.td_m.pkas[std_pka_level:]
            if deprot_steps > len(deprot_pkas):
                print(f'{self.id}: incomplete deprotonation due to insufficient pKa values.')
                deprot_steps = len(deprot_pkas)

            for pka in deprot_pkas[:deprot_steps]:
                dfg0_target += RT * np.log(10) * pka
                charge_target -= 1
                nh_target -= 1

        elif target_charge > self.td_m.charge_std:
            prot_steps = target_charge - self.td_m.charge_std
            prot_pkas = self.td_m.pkas[:std_pka_level]
            if prot_steps > len(prot_pkas):
                print(f'{self.id}: incomplete protonation due to insufficient pKa values.')
                prot_steps = len(prot_pkas)

            for pka in prot_pkas[-prot_steps:]:
                dfg0_target -= RT * np.log(10) * pka
                charge_target += 1
                nh_target += 1
        return dfg0_target, charge_target, nh_target

    def transform_std_gibbs_energy(self, dfg0_std, nh_std, charge_std):
        """Transform standard Gibbs energy of formation wrt compartmental pH and ionic strength.

        Alberty, 2003, equation 4.4-10:
            ∆fG'˚(I) = ∆fG˚(I=0) + RT ln(10) NH pH - RT ln(10) (z^2 - NH) A √I / (1 + B √I)

        for protons: dfg0_std = 0.0, nh_std = 1, charge = 1
        drf0_tr = RT np.log(10) compartment_ph

        :param float dfg0_std: Gibbs energy of formation in standard conditions (ionic strength of 0 M)
        :param int nh_std: number of Hydrogen atoms in the structure
        :param int charge_std: electrical charge
        :return: transformed gibbs energy of formation
        :rtype: float
        """
        ext_dh_factor = (DEBYE_HUCKEL_A * np.sqrt(self.td_c.ionic_str) /
                         (1.0 + DEBYE_HUCKEL_B * np.sqrt(self.td_c.ionic_str)))

        h_atom_adjustment = RT * np.log(10) * nh_std * self.td_c.ph
        ionic_str_adjustment = RT * np.log(10) * (charge_std ** 2 - nh_std) * ext_dh_factor

        return dfg0_std + h_atom_adjustment - ionic_str_adjustment

    def get_gibbs_formation_energy_std_tr(self):
        """Calculate the transformed Gibbs energy of formation for reactant (pseudoisomeric group).

        Thermodymamics of pseudoisomer groups at specified pH
        Equation 4.5-6 in Alberty's book:
            ∆fG'˚(I) = ∆fG1'˚(I) - RT ln(P(I))
            - P(I): binding polynomial (partitioning function) at specified ionic strength I
            - ∆fG1'˚(I): transformed Gibbs energy of formation of least protonated species of isomer group

        :returns: standard transformed Gibbs energy of formation in kJ/mol
        :rtype: float
        """

        # protons have a ∆fG`˚(I) of zero, see Alberty, 2003 eq. 4.4-10
        #   ∆fG`˚(I=0) is zero by default and ionic strength term is zero as well
        if self.td_metabolite != CPD_PROTON:
            # deprotonation steps from TD metabolite in standard state up to MAX_PH
            pka_levels_below_max_ph = sum([1 for pka in self.td_m.pkas if pka < MAX_PH])
            target_deprot_steps = pka_levels_below_max_ph - self.get_std_pka_level()

            # std formation energy of least protonated species, charge and number of H atoms in structure
            # Note: deprotonation may be incomplete due to insufficient protons or pKa values
            dfg0_lp, charge_lp, nh_lp = self.get_dfg0_at_charge(self.td_m.charge_std - target_deprot_steps)
        else:
            dfg0_lp = 0.0
            charge_lp = 1.0
            nh_lp = 1.0

        deprot_steps = self.td_m.charge_std - charge_lp

        # transform Gibbs formation energy of species wrt compartment pH and ionic strength
        dfg0_lp_tr = self.transform_std_gibbs_energy(dfg0_lp, nh_lp, charge_lp)

        # binding polynomial for group of species wrt compartment pH and ionic strength
        binding_polynomial, avg_h_binding, self.mole_fractions_rev = self._get_binding_polynomial(deprot_steps)

        # transformed energy of reactant wrt to compartment pH and I (Alberty, 2003, eq. 4.5-6)
        self.dfg0_tr = dfg0_lp_tr - RT * np.log(binding_polynomial)

        # average el. charge and average H atoms in the structures
        self.avg_h_atoms_tr = nh_lp + avg_h_binding
        self.avg_charge_tr = charge_lp + avg_h_binding
        return self.dfg0_tr

    def _transform_pkas_rev(self, deprot_steps):
        """Transform pKa values in pysiological range to given ionic strength

        pKa values have to be adapted to ionic strength so we can determine the
        binding polynomial. The pKa value become lower with increasing ionic strength.

        based on pyTFA MetaboliteThermo.get_pka(), but modified.

        Using Alberty, 2003, equation 3.6-6:
            ln K(I) = ln K(I=0) + A * ln(10) * √I/(1 + B√I) * ∑ v_i z_i^2
            ln K = ln(10) * log(K) = ln(10) * -pK
            pKa(I) = pKa(I=0) - A √I / (1 + B√I) * ∑ v_i z_i^2

            - A, B are Debye-Huckel parameters at 298.15K, 1 bar values (A = 0.510651, B = 1.6)
            - I is ionic strengh, z_i are charges, v_i are stoichiometric factors
            ∆rG'˚ = - RT ln(K')
                - K': apparent equilibrium constant
            pKa = -log10(K') = - ln(K')/ln(10)

            E.g. for dissociation reaction [HATP^(3-)] = [H+] + [ATP^(4-)]
            Acid dissociation - charges: (charge + 1) = 1 + charge
            ∑v_i z_i^2 = -(charge + 1)^2 + 1^2 + charge^2 = -charge^2  - 2 charge - 1 + 1 + charge^2 = -2 charge
            pKa(I) = pKa(I=0) + 2 * charge * A √I / (1 + B√I)

        :param int deprot_steps: deprotonation steps till maximum pH (>= 0)
        :returns: metabolite pkas in valid range in reverse order (highest to lowest)
        :rtype: list[float]
        """
        least_prot_charge = self.td_m.charge_std - deprot_steps
        ext_dh_factor = (DEBYE_HUCKEL_A * np.sqrt(self.td_c.ionic_str) /
                         (1.0 + DEBYE_HUCKEL_B * np.sqrt(self.td_c.ionic_str)))

        pkas_tr_rev = []
        for i, pka in enumerate(sorted([pka for pka in self.td_m.pkas if pka < MAX_PH], reverse=True)):
            pka_tr = pka + 2 * (least_prot_charge + i) * ext_dh_factor
            pkas_tr_rev.append(pka_tr)
        return pkas_tr_rev

    def _get_binding_polynomial(self, deprot_steps):
        """Calculate the transformed binding polynomial of a reactant and related parameters.

        Requried for reactants consisting of multiple species with
        different protonation states in the pysiological pH range.

        binding polynomial
        Alberty, 2003, equation 4.5-7:
            P = 1 + [H+]/K1 + [H+]^2/K1K2 +
            - K1, K2, ...: successive acid dissociation constants at the specified pH
            and ionic strength, starting with the highest pK
            - Kx = 10**(-pKax)
            - [H+] = 10**(-pH)

        average binding of hydrogen ion avg_NH
        Alberty, 2003, equations 1.3-7, -8, -9
        - avg_NH = [H+]/P dP/d[H+]                       (1.3-9)
        - P = 1 + [H+]/K1 + [H+]^2/K1K2 +                (1.3-8)
        - r_1 = 1/P, r_2 = [H+]/K1, r_3 = [H+]^2/K1K2    (1.3-3, 1.3-4, 1.3-5)
        - N = [H+] dP/d[H+] = [H+]/K1 + 2*[H+]^2/K1K2 +
        - NH = N/P = ([H+]/K1 + 2*[H+]^2/K1K2 + )/(1 + [H+]/K1 + [H+]^2/K1K2 + )   (1.3-7)

        :param int deprot_steps: deprotonation steps till maximum pH (>= 0)
        :returns: binding polynomial, averge hydrogen ion binding, mole fractions (reverse)
        :rtype: float, float, numpy.ndarray
        """
        pkas_tr_rev = self._transform_pkas_rev(deprot_steps)

        proton_conc = 10.0 ** -self.td_c.ph
        prod_eq_constants = 1.0
        factors = [1.0]
        polynomial = 1.0
        for i, pka_tr in enumerate(pkas_tr_rev):
            prod_eq_constants *= 10.0 ** (-pka_tr)
            factor = proton_conc ** (i + 1) / prod_eq_constants
            polynomial += factor
            factors.append(factor)

        avg_h_binding = sum([i * factor for i, factor in enumerate(factors)]) / polynomial
        mole_fractions_rev = np.array([factor / polynomial for factor in factors])

        return polynomial, avg_h_binding, mole_fractions_rev

    def modify_attribute(self, attribute, value):
        """modify attribute value.

        :param str attribute: attribute name
        :param value: value to be configured
        :type value: str, float, int
        """
        if hasattr(self, attribute):
            setattr(self, attribute, value)
        else:
            print(f'unknown TD reactant attribute {attribute}')



"""Implementation of EcmOptimization class.

Support functions for COBRApy and gurobipy optimization of EC (enzyme constraint) models
that have been created using the f2xba modelling package.

Support of GECKO, ccFBA, MOMENTmr and MOMENT optimization, as well
support for thermodynamic enhance models (TGECKO, TccFBA, TMOMENTmr, TMOMENT)

Peter Schubert, HHU Duesseldorf, CCB, April 2024
"""

import re
import pandas as pd
import tqdm
from collections import defaultdict
# gurobipy should not be a hard requirement, unless used in this context
try:
    import gurobipy as gp
except ImportError:
    gp = None
    pass

import f2xba.prefixes as pf
from .optimize import Optimize
from .ecm_results import EcmResults


class EcmOptimization(Optimize):
    """Optimization support for enzyme constraint models (ECM) like GECKO.

    Using the gurobipy interface requires that GUROBI and gurobipy being installed on your system.

    Ref: Sánchez, B. J., Zhang, C., Nilsson, A., Lahtvee, P. J., Kerkhoven, E. J., & Nielsen, J. (2017).
    Improving the phenotype predictions of a yeast genome‐scale metabolic model by incorporating enzymatic
    constraints. Molecular Systems Biology, 13(8), 935. https://doi.org/https://doi.org/10.15252/msb.20167411

    Using the gurobipy interface requires that GUROBI and gurobipy being installed on your system:


    When using this class with COBRApy instead, supply the reference of the COBRApy model during instantiation.
    While strictly, this class is not required when using the COBRApy interface, it provides access to
    features implemented in f2xba, like optimization results analysis.

    Using the gurobipy interface for optimization of enzyme constraint models, like GECKO or TGECKO.
    Note: GUROBI optimizer with gurobipy (https://www.gurobi.com) needs to be installed on your system.

    Using the COBRApy interface for optimization of enzyme constraint models, like GECKO or TGECKO:
    Use of EcmOptimization is optional, but it can provide access to additional features, e.g. results analysis,
    and ensures correct configuration of variables and constraints for thermodynamics constraint variants.

    .. code-block:: python

        import cobrapy

        cobra_model = cobra.io.read_sbml_model('iML1515_GECKO.xml')
        eo = EcmOptimization('iML1515_GECKO.xml', cobra_model)
        cobra_model.medium = {rid: 1000.0 for rid in lb_medium}
        solution = cobra_model.optimize()


    Using the gurobipy interface for optimization of enzyme constraint models, like GECKO or TGECKO:
    Note: GUROBI optimizer with gurobipy (https://www.gurobi.com) needs to be installed on your system.

    .. code-block:: python

        eo = EcmOptimization('iML1515_GECKO.xml')
        eo.medium = {rid: 1000.0 for rid in lb_medium}
        solution = eo.optimize()
    """

    def __init__(self, fname, cobra_model=None):
        """Instantiate the EcmOptimization instance.

        :param str fname: filename of the SBML coded extended model
        :param cobra_model: reference to a COBRApy model (default: None)
        :type cobra_model: :class:`cobra.Model`
        """
        super().__init__('ECM', fname, cobra_model)
        self.orig_coupling = {}

        self.ecm_type = self.m_dict['modelAttrs'].get('id', '_GECKO').rsplit('_', 1)[1]
        if self.ecm_type.endswith('MOMENT'):
            self._configure_moment_model_constraints()

    @property
    def medium(self):
        """Mimic medium property of COBRApy to set and retrieve medium.

        :return: exchange reaction ids with (positive valued) uptake rates
        :rtype: dict
        """
        ex_medium = {}
        for ex_rid, (lb, ub) in self.get_variable_bounds(self.uptake_rids).items():
            if lb < 0.0:
                ex_medium[re.sub('^R_', '', ex_rid)] = -lb
        return ex_medium

    @medium.setter
    def medium(self, ex_medium):
        """Mimic medium property of COBRApy for medium assignments

        :param dict(str, float) ex_medium: exchange reaction ids (with/without `R_`) and uptake rates
        """
        self.set_medium(ex_medium)

    def _configure_moment_model_constraints(self):
        """Configure constraints related to MOMENT model optimization.

        Must be invoked after loading a MOMENT type model (not MOMENTmr). MOMENT implements promiscuous
        enzymes, that can catalyze alternative reactions without additional costs. The most costly
        reaction flux needs to be supported and all alternative reaction fluxes come for free.
        """
        if self.is_gpm:
            for constr in self.gpm.getConstrs():
                if re.match(pf.C_prot_, constr.ConstrName) and re.match(pf.C_prot_pool, constr.ConstrName) is None:
                    constr.sense = '>'
        else:
            for constr in self.model.constraints:
                if re.match(pf.C_prot_, constr.name) and re.match(pf.C_prot_pool, constr.name) is None:
                    constr.ub = 1000.0
        print(f'MOMENT protein constraints configured ≥ 0')

    def scale_kcats(self, scale_kcats):
        """Scale turnover numbers in GECKO models to support manual parameter tuning.

        Use unscale_kcats() to return to original values.

        .. code-block:: python

            eo = EcmOptimization('iML1515_GECKO.xml)
            scale_kcats= {'BPNT': 0.2, 'TMPK': 0.2, 'TMPK_REV': 0.2, 'CLPNS': 0.25}
            eo.scale_kcats(scale_kcats)
            solution = eo.optimize()
            eo.unscale_kcats()

        :param dict(str, float) scale_kcats: reaction ids (without prefix `R_`) with scaling factor
        """
        if self.is_gpm:
            self._gp_scale_kcats(scale_kcats)
        else:
            self._cp_scale_kcats(scale_kcats)

    def unscale_kcats(self):
        """Reset previously scaled turnover numbers.

        see scale_kcats().
        """
        if self.is_gpm:
            self._gp_unscale_kcats()
        else:
            self._cp_unscale_kcats()

    def _cp_scale_kcats(self, scale_kcats):
        """Scale kcat values for COBRApy interface, by updating coupling coefficients.

        Reaction id without 'R_' prefix and kcat scaling factors.

        :param dict(str, float) scale_kcats: selected reaction ids (without 'R_' prefix) with kcat scaling factor
        """
        assert self.is_gpm is False, 'applicable to COBRApy interface only'
        orig_coupling = defaultdict(dict)
        for ridx, scale in scale_kcats.items():
            if ridx in self.model.reactions:
                r = self.model.reactions.get_by_id(ridx)
                for m, coeff in r.metabolites.items():
                    if re.match(f'{pf.C_prot_}', m.id):
                        r.add_metabolites({m: coeff / scale}, combine=False)
                        orig_coupling[ridx][m.id] = coeff
            else:
                print(f'Enzyme constraint variable not found for reaction {ridx}')
        self.orig_coupling = dict(orig_coupling)

    def _cp_unscale_kcats(self):
        """Unscale kcat values for COBRApy interface, by resetting original coupling coefficients
        """
        assert self.is_gpm is False, 'applicable to COBRApy interface only'
        for ridx, couplings in self.orig_coupling.items():
            r = self.model.reactions.get_by_id(ridx)
            for constr_id, coeff in couplings.items():
                r.add_metabolites({constr_id: coeff}, combine=False)
        self.orig_coupling = {}

    def _gp_scale_kcats(self, scale_kcats):
        """Scale kcat values for gurobipy interface, by updating coupling coefficients.

        :param dict(str,float) scale_kcats: selected reaction ids (without 'R_' prefix) with kcat scaling factor
        """
        assert self.is_gpm is True, 'applicable to gurobipy interface only'
        orig_coupling = defaultdict(dict)
        self.gpm.update()
        for ridx, scale in scale_kcats.items():
            var = self.gpm.getVarByName(f'{pf.R_}{ridx}')
            if var:
                col = self.gpm.getCol(var)
                for idx in range(col.size()):
                    constr = col.getConstr(idx)
                    if re.match(f'{pf.C_prot_}', constr.getAttr('ConstrName')):
                        coeff = col.getCoeff(idx)
                        self.gpm.chgCoeff(constr, var, coeff / scale)
                        orig_coupling[ridx][constr.getAttr('ConstrName')] = coeff
            else:
                print(f'Enzyme constraint variable not found for reaction {ridx}')

        self.gpm.update()
        self.orig_coupling = dict(orig_coupling)

    def _gp_unscale_kcats(self):
        """reset kcat values to original values.

        Used in manually tuning model kcats
        - call scale_kcats(scale_kcats) prior to optmization
        - call unscale_kcats() after optimization to reset old kcat values
        """
        assert self.is_gpm is True, 'applicable to gurobipy interface only'
        for ridx, couplings in self.orig_coupling.items():
            var = self.gpm.getVarByName(f'{pf.R_}{ridx}')
            for constr_id, coeff in couplings.items():
                constr = self.gpm.getConstrByName(constr_id)
                self.gpm.chgCoeff(constr, var, coeff)
        self.gpm.update()
        self.orig_coeffs = {}

    def single_gene_deletion(self, genes=None, method='ecm', solution=None, **kwargs):
        """Perform a single gene deletion analysis for enzyme constraint models using gurobipy interface.

        Interface aligned to COBRApy single_gene_deletion() method.
        Perform single gene deletion simulations for provided list of gene in `genes`.
        If `genes` is not provided, perform gene deletion for all genes that may be
        active in the wild type solution. In case a gene is not required in the wild type solution, a knockout
        simulation is not performed for this gene. Its growth rate value is set to the wild type value and
        its optimization status is set to `wt_solution`.
        A wild type solution can be provided with parameter `solution`, alternatively a wild type solution
        is determined automatically.

        When `method` is set to `room` or `linear room`, following keyword arguments can be added:
        `delta`: relative tolerance range (default: 0.03),
        `epsilon`: absolute tolerance range (default: 1e-3),
        `time_limit`: in seconds for single gene deletion simulation, used for 'room' (default: 30.0).

        .. code-block:: python

            eo = EcmOptimization('iML1515_GECKO.xml')
            eo.medium = {rid: 1000.0 for rid in lb_medium}
            df_sgko = eo.single_gene_deletion()

        Example for MOMA based SGKO analysis for selected genes with wild type solution provided:

        .. code-block:: python

            wt_solution = eo.optimize()
            eo.single_gene_deletion(genes = ['b0002', 'b0003', 'b0007', 'b0025'], method='moma', solution=wt_solution)

        :param list or set genes: (optional) gene ids
        :param str method: (optional) alternative methods 'moma', 'linear moma', 'room' or 'linear room'
        :param solution: (optional) wild type ECM solution
        :type solution: :class:`Solution`
        :param kwargs: keyword arguments passed on to 'room' and 'linear room' methods
        :return: table with SGKO results, containing growth rate in h-1, optimization status and fitness
        :rtype: pandas.DataFrame
        """

        if self.is_gpm is None:
            print('Method implemented for gurobipy interface only.')
            return pd.DataFrame()

        linear = True if 'linear' in method else False
        biomass_rid = self._get_biomass_rid()

        # determine wild type growth rate and fluxes, if solution not provided
        # extract wt_gr from fluxes of biomass_rid (assume growth maximization of Biomass objective function)
        if solution is None:
            solution = self.optimize()
        wt_gr = solution.fluxes[biomass_rid]
        wt_fluxes = dict(EcmResults(self, {'wt': solution}).collect_fluxes()['wt'])

        # determine gene list, if not provided
        if genes is None:
            genes = self.get_active_genes(wt_fluxes)
            all_genes = self.m_dict['fbcGeneProducts'].label.values
        else:
            all_genes = genes

        # optimization loop for single gene deletions
        cols = ['growth_rate', 'status', 'fitness']
        sgko_results = {'wt': [wt_gr, 'wt_solution', 1.0]}
        for gene in tqdm.tqdm(sorted(all_genes)):
            if gene in genes:

                # simulate a single gene deletion
                orig_rid_bounds = self.gene_knock_outs(gene)
                if 'moma' in method:
                    solution = self.moma(wt_fluxes, linear=linear)
                elif 'room' in method:
                    solution = self.room(wt_fluxes, linear=linear, **kwargs)
                else:
                    solution = self.optimize()
                self.set_variable_bounds(orig_rid_bounds)

                # process simulation result
                if solution is None:
                    sgko_results[gene] = [0.0, 'infeasible', 0.0]
                elif solution.status in {'optimal', 'time_limit', 'suboptimal'}:
                    if 'moma' in method or 'room' in method:
                        mutant_gr = solution.fluxes[biomass_rid]
                    else:
                        mutant_gr = solution.objective_value
                    sgko_results[gene] = [mutant_gr, solution.status, mutant_gr / wt_gr]
                else:
                    sgko_results[gene] = [0.0, solution.status, 0.0]
            else:
                # genes not in gene_list are assumed to not impact the wild type solution
                sgko_results[gene] = [wt_gr, 'wt_solution', 1.0]

        df_sgko = pd.DataFrame(sgko_results.values(), index=list(sgko_results), columns=cols)
        df_sgko.index.name = 'gene'

        return df_sgko


"""Implementation of BiocycData class.

Peter Schubert, CCB, HHU Duesseldorf, November 2022
"""

import os
import pandas as pd
import urllib.parse
import urllib.request

from .biocyc_gene import BiocycGene
from .biocyc_protein import BiocycProtein
from .biocyc_rna import BiocycRNA

api_url = 'https://websvc.biocyc.org/xmlquery'


class BiocycData:
    """Access to enzyme related data from BioCyc online database.

    Access BioCyc data via BioVelo query, see https://biocyc.org/web-services.shtml

    Enzyme constraint and resource balance constraint models require information on enzyme composition
    during model construction. By default, enzymes will be composed of one copy per gene product, derived
    for the reaction gene product reaction rule configuration of the SBML model. Alternatively, enzyme
    composition can be retrieved from BioCyc online resource or loaded from an
    enzyme composition configuration file.

    Enzyme composition derived from BioCyc would be suitable for highly curated organism databases,
    like E. coli K-12. Access to BioCyc resources requires a paid BioCyc subscription, depending on organism.

    Initially, enzyme composition data could be retrieved from BioCyc and exported to file
    using XbaModel.export_enz_composition(). The enzyme composition could be manually adjusted and
    used for subsequent model creations.

    The organism in the BioCyc database is referenced by `org_prefix`.
    In order to use enzyme composition data from Biocyc, use configuration data in the XBA configuration file,
    sheet `general`. The parameter `biocyc_org_prefix` references the organism in BioCyc, the parameter
    `organism_dir` specifies the local download directory.
    Delete locally stored BioCyc files to enforce a retrieval from the online database.
    Enzyme composition can be loaded from file by configuring the parameter `enzyme_comp_fname` in the
    XBA parameter file, sheet `general`.

    Example: Access enzyme configuration for E. coli K-12 MG1655 strain (org_prefix: ecoli).

        .. code-block:: python

            from f2xba.biocyc.biocyc_data import BiocycData

            biocyc_data = BiocycData('ecoli', 'data_refs/biocyc')

            gene = 'b0928'
            bc_gene = biocyc_data.locus2gene[gene]
            bc_protein = biocyc_data.genes[bc_gene].proteins[0]
            biocyc_data.proteins[bc_protein].__dict__

    :param str org_prefix: BioCyc organism reference
    :param str biocyc_dir: directory name, where downloads of BioCyc are stored
    """

    def __init__(self, org_prefix, biocyc_dir):
        """Instantiate BiocycData Instance

        For a given organism, query BioCyc on-line database for specific components in
         suitable detail level. Subsequently, extract relevant information.
        Use already downloaded BioCyc exports in case they exist in biocyc_dir.

        :param str org_prefix: BioCyc organism reference
        :param str biocyc_dir: directory name, where downloads of BioCyc are stored
        """

        self.prefix_lc = org_prefix.lower()
        self.prefix_uc = org_prefix.upper() + ':'
        self.url = api_url
        self.biocyc_dir = biocyc_dir

        # Exports with required level of detail.
        self.biocyc_data = {'Gene': ['genes', 'full'],
                            'Protein': ['proteins', 'low'],
                            'RNA': ['RNAs', 'low']}

        if not self._is_complete_biocyc_data():
            self._retrieve_biocyc_data()

        self.genes = BiocycGene.get_genes(self._biocyc_data_fname('Gene'))
        """BioCyc gene related data, referenced by BioCyc gene id."""

        self.locus2gene = {gene.locus: bc_id for bc_id, gene in self.genes.items()}
        """Map gene id (locus) to BioCyc gene id."""

        self.proteins = BiocycProtein.get_proteins(self._biocyc_data_fname('Protein'))
        """BioCyc protein related data, referenced by BioCyc protein id."""

        self.rnas = BiocycRNA.get_rnas(self._biocyc_data_fname('RNA'))
        """BioCyc RNA related data, referenced by BioCyc RNA id."""

        # set gene locus on direct gene product and rnas
        for protein in self.proteins.values():
            if type(protein.gene) is str and protein.gene in self.genes:
                protein.gene_composition = {self.genes[protein.gene].locus: 1.0}
        for rna in self.rnas.values():
            if type(rna.gene) is str and rna.gene in self.genes:
                rna.gene_composition = {self.genes[rna.gene].locus: 1.0}

        # iteratively configure gene composition
        for protein_id, p in self.proteins.items():
            if len(p.protein_parts) > 0:
                p.gene_composition = self.get_gene_composition(protein_id)

    @staticmethod
    def _add_composition(tot_composition, composition, p_stoic):
        for item, stoic in composition.items():
            if item not in tot_composition:
                tot_composition[item] = 0
            tot_composition[item] += stoic * p_stoic
        return tot_composition

    def add_simple_proteins(self, add_proteins):
        """Add simple proteins (i.e. direct gene products).

        Proteins have a unique Protein id, e.g. 'GATC-MONOMER'
        and parameters, supplied in a dict
        Protein instance is created, configured and added to dict of
        existing proteins in BioCyc model.
        Gene needs to exist.
        Gene is identified by gene locus, e.g. 'b2092'
        Gene record is updated with new protein id

        :meta private:
        :param dict(str, dict) add_proteins: proteins with configuration data
        :return: number of added proteins
        :rtype: int
        """
        n_updates = 0
        for pid, data in add_proteins.items():
            if 'locus' in data:
                gid = self.locus2gene[data['locus']]
                p = BiocycProtein(pid)
                p.gene = gid
                p.gene_composition = {data['locus']: 1.0}
                self.proteins[pid] = p
                self.modify_proteins({pid: data})
                if pid not in self.genes[gid].proteins:
                    self.genes[gid].proteins.append(pid)
                n_updates += 1
        return n_updates

    def modify_proteins(self, protein_updates):
        """Modify configuration of proteins.

        Proteins have a unique Protein id, e.g. 'CPLX0-231'

        :meta private:
        :param dict(str, dict) protein_updates: proteins with configuration data
        :return: number of added proteins
        :rtype: int
        """
        n_updates = 0
        for pid, data in protein_updates.items():
            if pid in self.proteins:
                p = self.proteins[pid]
                for key, val in data.items():
                    if hasattr(p, key):
                        setattr(p, key, val)
                n_updates += 1
        return n_updates

    def get_gene_composition(self, protein_id):
        """Retrieve gene composition of an enzyme (BioCyc protein).

        .. code-block:: python

            biocyc_data.get_gene_composition('ASPAMINOTRANS-MONOMER')

        :param str protein_id: BioCyc protein identifier
        :return: gene composition of enzyme
        :rtype: dict(str, float)
        """
        gene_composition = {}
        p = self.proteins[protein_id]

        if len(p.gene_composition) > 0:
            # in case a protein is a direct gene product, its composition contains the gene reference.
            # create new dict with gene composition to avoid side effects
            gene_composition = {gene: stoic for gene, stoic in p.gene_composition.items()}
        else:
            # retrieve composition information for a protein complex (iteratively)
            for p_part_id, p_stoic in p.protein_parts.items():
                composition = self.get_gene_composition(p_part_id)
                gene_composition = self._add_composition(gene_composition, composition, p_stoic)
            # add rna gene composition
            for rna_id, p_stoic in p.rna_parts.items():
                composition = self.rnas[rna_id].gene_composition
                gene_composition = self._add_composition(gene_composition, composition, p_stoic)
        return gene_composition

    def _biocyc_data_fname(self, component):
        class_name, detail = self.biocyc_data[component]
        return os.path.join(self.biocyc_dir, f'biocyc_{class_name}_{detail}.xml')

    def _is_complete_biocyc_data(self):
        exports_available = True
        for component in self.biocyc_data:
            file_name = self._biocyc_data_fname(component)
            if not os.path.exists(file_name):
                exports_available = False
                print(f'{file_name} does not exist.')
        return exports_available

    def _retrieve_biocyc_data(self):
        """Retrieve data from BioCyc site using REST API

        using BioVelo query, see https://biocyc.org/web-services.shtml
        """
        for component in self.biocyc_data:
            file_name = self._biocyc_data_fname(component)
            if not os.path.exists(file_name):
                class_name, detail = self.biocyc_data[component]
                print('retrieve ' + class_name + ' from BioCyc at level ' + detail)
                base_url = urllib.parse.urlparse(self.url)
                biovelo_qry = f'[x: x<-{self.prefix_lc}^^{class_name}]'
                query = urllib.parse.quote(biovelo_qry) + '&detail=' + detail
                split_url = base_url._replace(query=query)
                url = urllib.parse.urlunparse(split_url)
                req = urllib.request.Request(url)

                with urllib.request.urlopen(req) as response, open(file_name, 'w') as file:
                    file.write(response.read().decode('utf-8'))
                print(f'{file_name} from BioCyc retrieved')

    def _get_protein_components(self, protein_id):
        """Determine components of a given protein/enzyme.

        recursively parse through proteins and extract compounds
        and loci for proteins, RNAs. All with stoichiometry
        """
        components = {'proteins': {}, 'rnas': {}}   # 'compounds': {}}
        gene = self.proteins[protein_id].gene
        if gene is not None:
            locus = self.genes[gene].locus
            components['proteins'] = {locus: 1.0}
        else:
            for rna_part in self.proteins[protein_id].rna_parts:
                rna, stoic_str = rna_part.split(':')
                gene = self.rnas[rna].gene
                locus = self.genes[gene].locus
                stoic = int(stoic_str)
                if locus not in components['rnas']:
                    components['rnas'][locus] = stoic
                else:
                    components['rnas'][locus] += stoic
            for protein_part in self.proteins[protein_id].protein_parts:
                protein, stoic_str = protein_part.split(':')
                stoic = int(stoic_str)
                sub_components = self._get_protein_components(protein)
                # update components:
                for part_type in sub_components:
                    for component in sub_components[part_type]:
                        if component not in components[part_type]:
                            components[part_type][component] = stoic * sub_components[part_type][component]
                        else:
                            components[part_type][component] += stoic * sub_components[part_type][component]
        return components

    def export_enzyme_composition(self, fname):
        """Export BioCyc enzyme composition to Excel spreadsheet.

        .. code-block:: python

            biocyc_data.export_enzyme_composition('BioCyc_enz_composition.xlsx')

        :param str fname: file name of export file with extension '.xlsx'
        """
        enz_comp = {}
        for enz_id, enz in self.proteins.items():
            if len(enz.gene_composition) > 0 and len(enz.enzrxns) > 0:
                gene_comp = '; '.join([f'gene={gene}, stoic={stoic}'
                                       for gene, stoic in enz.gene_composition.items()])
                enz_comp[enz_id] = [enz.name, enz.synonyms, gene_comp]

        df_enz_comp = pd.DataFrame(enz_comp.values(), index=list(enz_comp), columns=['name', 'synonyms', 'genes'])
        df_enz_comp.index.name = 'biocyc_id'

        with pd.ExcelWriter(fname) as writer:
            df_enz_comp.to_excel(writer, sheet_name='composition')
            print(f'{len(df_enz_comp)} enzyme compositions written to {fname}')
